from __future__ import annotations

import enum
import uuid
from abc import ABCMeta, abstractmethod
from collections.abc import Container, Iterable, Mapping, Sequence
from dataclasses import dataclass, field
from typing import Any, Callable, Generic, Optional, Self, TypeAlias, TypeVar, cast

import sqlalchemy as sa
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import load_only, selectinload, with_loader_criteria

from .context import ClientContext
from .exceptions import InvalidScope, ScopeTypeMismatch
from .permission_defs import BasePermission

__all__: Sequence[str] = (
    "DomainScope",
    "ProjectScope",
    "UserScope",
    "StorageHost",
    "ImageRegistry",
    "ScalingGroup",
    "AbstractPermissionContext",
    "AbstractPermissionContextBuilder",
)


PermissionType = TypeVar("PermissionType", bound=BasePermission)


class PredefinedRole(enum.StrEnum):
    OWNER = enum.auto()
    ADMIN = enum.auto()
    MONITOR = enum.auto()
    # Privileged member is a user who is part of a specific scope
    # and has read(or some additional permissions) to objects that the scope has.
    # e.g. Project member
    PRIVILEGED_MEMBER = enum.auto()
    # Member is a user who is part of a specific scope
    # and has NO PERMISSION to objects that the scope has.
    # e.g. Domain member
    MEMBER = enum.auto()


_EMPTY_FSET: frozenset = frozenset()


async def get_predefined_roles_in_scope(
    ctx: ClientContext,
    scope: ScopeType,
    db_session: AsyncSession | None = None,
) -> frozenset[PredefinedRole]:
    from ..user import UserRole

    async def _calculate_role(db_session: AsyncSession) -> frozenset[PredefinedRole]:
        match ctx.user_role:
            case UserRole.SUPERADMIN:
                return await _calculate_role_in_scope_for_suadmin(ctx, db_session, scope)
            case UserRole.MONITOR:
                return await _calculate_role_in_scope_for_monitor(ctx, db_session, scope)
            case UserRole.ADMIN:
                return await _calculate_role_in_scope_for_admin(ctx, db_session, scope)
            case UserRole.USER:
                return await _calculate_role_in_scope_for_user(ctx, db_session, scope)

    if db_session is None:
        async with ctx.db.begin_readonly_session() as _db_session:
            return await _calculate_role(_db_session)
    else:
        return await _calculate_role(db_session)


async def _calculate_role_in_scope_for_suadmin(
    ctx: ClientContext, db_session: AsyncSession, scope: ScopeType
) -> frozenset[PredefinedRole]:
    from ..domain import DomainRow
    from ..group import AssocGroupUserRow, GroupRow
    from ..user import UserRow

    match scope:
        case SystemScope():
            return frozenset([PredefinedRole.ADMIN])
        case DomainScope(domain_name):
            stmt = (
                sa.select(DomainRow)
                .where(DomainRow.name == domain_name)
                .options(load_only(DomainRow.name))
            )
            domain_row = cast(DomainRow | None, await db_session.scalar(stmt))
            if domain_row is not None:
                return frozenset([PredefinedRole.ADMIN])
            else:
                return _EMPTY_FSET
        case ProjectScope(project_id):
            stmt = (
                sa.select(GroupRow)
                .where(GroupRow.id == project_id)
                .options(
                    selectinload(GroupRow.users),
                    with_loader_criteria(
                        AssocGroupUserRow, AssocGroupUserRow.user_id == ctx.user_id
                    ),
                )
            )
            project_row = cast(GroupRow | None, await db_session.scalar(stmt))
            if project_row is None:
                return _EMPTY_FSET
            result = frozenset([PredefinedRole.ADMIN])
            if project_row.users:
                result = frozenset([PredefinedRole.OWNER])
            return result
        case UserScope(user_id):
            if ctx.user_id == user_id:
                return frozenset([PredefinedRole.OWNER])
            stmt = (
                sa.select(UserRow).where(UserRow.uuid == user_id).options(load_only(UserRow.uuid))
            )
            user_row = cast(UserRow | None, await db_session.scalar(stmt))
            if user_row is not None:
                return frozenset([PredefinedRole.ADMIN])
            else:
                return _EMPTY_FSET


async def _calculate_role_in_scope_for_monitor(
    ctx: ClientContext, db_session: AsyncSession, scope: ScopeType
) -> frozenset[PredefinedRole]:
    from ..domain import DomainRow
    from ..group import AssocGroupUserRow, GroupRow
    from ..user import UserRow

    match scope:
        case SystemScope():
            return frozenset([PredefinedRole.MONITOR])
        case DomainScope(domain_name):
            stmt = (
                sa.select(DomainRow)
                .where(DomainRow.name == domain_name)
                .options(load_only(DomainRow.name))
            )
            domain_row = cast(DomainRow | None, await db_session.scalar(stmt))
            if domain_row is not None:
                return frozenset([PredefinedRole.MONITOR])
            else:
                return _EMPTY_FSET
        case ProjectScope(project_id):
            stmt = (
                sa.select(GroupRow)
                .where(GroupRow.id == project_id)
                .options(
                    load_only(GroupRow.id, GroupRow.domain_name),
                    selectinload(GroupRow.users),
                    with_loader_criteria(
                        AssocGroupUserRow, AssocGroupUserRow.user_id == ctx.user_id
                    ),
                )
            )
            project_row = cast(GroupRow | None, await db_session.scalar(stmt))
            if project_row is None:
                return _EMPTY_FSET
            if project_row.domain_name == ctx.domain_name:
                result = frozenset([PredefinedRole.ADMIN])
            else:
                return _EMPTY_FSET
            if project_row.users:
                result = frozenset([*result, PredefinedRole.PRIVILEGED_MEMBER])
            return result
        case UserScope(user_id):
            if ctx.user_id == user_id:
                return frozenset([PredefinedRole.OWNER])
            stmt = (
                sa.select(UserRow).where(UserRow.uuid == user_id).options(load_only(UserRow.uuid))
            )
            user_row = cast(UserRow | None, await db_session.scalar(stmt))
            if user_row is not None:
                return frozenset([PredefinedRole.MONITOR])
            else:
                return _EMPTY_FSET


async def _calculate_role_in_scope_for_admin(
    ctx: ClientContext, db_session: AsyncSession, scope: ScopeType
) -> frozenset[PredefinedRole]:
    from ..group import AssocGroupUserRow, GroupRow
    from ..user import UserRow

    match scope:
        case SystemScope():
            return frozenset([PredefinedRole.ADMIN])
        case DomainScope(domain_name):
            if ctx.domain_name == domain_name:
                return frozenset([PredefinedRole.ADMIN])
            else:
                return _EMPTY_FSET
        case ProjectScope(project_id):
            stmt = (
                sa.select(GroupRow)
                .where(GroupRow.id == project_id)
                .options(
                    load_only(GroupRow.id, GroupRow.domain_name),
                    selectinload(GroupRow.users),
                    with_loader_criteria(
                        AssocGroupUserRow, AssocGroupUserRow.user_id == ctx.user_id
                    ),
                )
            )
            project_row = cast(GroupRow | None, await db_session.scalar(stmt))
            if project_row is None:
                return _EMPTY_FSET

            if project_row.domain_name == ctx.domain_name:
                result = frozenset([PredefinedRole.ADMIN])
            else:
                return _EMPTY_FSET
            if project_row.users:
                result = frozenset([PredefinedRole.OWNER])
            return result
        case UserScope(user_id, domain_name):
            if ctx.user_id == user_id:
                return frozenset([PredefinedRole.OWNER])
            if domain_name is not None:
                _domain_name = domain_name
            else:
                stmt = (
                    sa.select(UserRow)
                    .where(UserRow.uuid == user_id)
                    .options(load_only(UserRow.domain_name))
                )
                user_row = cast(UserRow | None, await db_session.scalar(stmt))
                if user_row is None:
                    return _EMPTY_FSET
                _domain_name = user_row.domain_name
            if _domain_name == ctx.domain_name:
                return frozenset([PredefinedRole.ADMIN])
            else:
                return _EMPTY_FSET


async def _calculate_role_in_scope_for_user(
    ctx: ClientContext, db_session: AsyncSession, scope: ScopeType
) -> frozenset[PredefinedRole]:
    from ..group import AssocGroupUserRow

    match scope:
        case SystemScope():
            return frozenset([PredefinedRole.MEMBER])
        case DomainScope(domain_name):
            if ctx.domain_name == domain_name:
                return frozenset([PredefinedRole.MEMBER])
            else:
                return _EMPTY_FSET
        case ProjectScope(project_id):
            stmt = (
                sa.select(AssocGroupUserRow)
                .where(
                    (AssocGroupUserRow.user_id == ctx.user_id)
                    & (AssocGroupUserRow.group_id == project_id)
                )
                .options(load_only(AssocGroupUserRow.user_id))
            )
            assoc_row = cast(AssocGroupUserRow | None, await db_session.scalar(stmt))
            if assoc_row is not None:
                return frozenset([PredefinedRole.PRIVILEGED_MEMBER])
            else:
                return _EMPTY_FSET
        case UserScope(user_id):
            if ctx.user_id == user_id:
                return frozenset([PredefinedRole.OWNER])
            else:
                return _EMPTY_FSET


class BaseScope(metaclass=ABCMeta):
    @abstractmethod
    def __str__(self) -> str:
        pass

    @abstractmethod
    def serialize(self) -> str:
        pass

    @classmethod
    @abstractmethod
    def deserialize(cls, val: str) -> Self:
        pass


@dataclass(frozen=True)
class SystemScope(BaseScope):
    def __str__(self) -> str:
        return "system scope()"

    def serialize(self) -> str:
        return "system:"

    @classmethod
    def deserialize(cls, val: str) -> Self:
        type_, _, _ = val.partition(":")
        if type_ != "system":
            raise ScopeTypeMismatch
        return cls()


@dataclass(frozen=True)
class DomainScope(BaseScope):
    domain_name: str

    def __str__(self) -> str:
        return f"Domain(name: {self.domain_name})"

    def serialize(self) -> str:
        return f"domain:{self.domain_name}"

    @classmethod
    def deserialize(cls, val: str) -> Self:
        type_, _, domain_name = val.partition(":")
        if type_ != "domain":
            raise ScopeTypeMismatch
        return cls(domain_name)


@dataclass(frozen=True)
class ProjectScope(BaseScope):
    project_id: uuid.UUID
    domain_name: str | None = None

    def __str__(self) -> str:
        return f"Project(id: {self.project_id}, domain: {self.domain_name}])"

    def serialize(self) -> str:
        if self.domain_name is not None:
            return f"project:{self.project_id}:{self.domain_name}"
        else:
            return f"project:{self.project_id}"

    @classmethod
    def deserialize(cls, val: str) -> Self:
        type_, _, values = val.partition(":")
        if type_ != "project":
            raise ScopeTypeMismatch
        raw_project_id, _, domain_name = values.partition(":")
        if domain_name:
            return cls(uuid.UUID(raw_project_id), domain_name)
        else:
            return cls(uuid.UUID(raw_project_id))


@dataclass(frozen=True)
class UserScope(BaseScope):
    user_id: uuid.UUID
    domain_name: str | None = None

    def __str__(self) -> str:
        return f"User(id: {self.user_id}, domain: {self.domain_name})"

    def serialize(self) -> str:
        if self.domain_name is not None:
            return f"user:{self.user_id}:{self.domain_name}"
        else:
            return f"user:{self.user_id}"

    @classmethod
    def deserialize(cls, val: str) -> Self:
        type_, _, values = val.partition(":")
        if type_ != "user":
            raise ScopeTypeMismatch
        raw_user_id, _, domain_name = values.partition(":")
        if domain_name:
            return cls(uuid.UUID(raw_user_id), domain_name)
        else:
            return cls(uuid.UUID(raw_user_id))


ScopeType: TypeAlias = SystemScope | DomainScope | ProjectScope | UserScope


def deserialize_scope(val: str) -> ScopeType:
    """
    Deserialize a string value in the format '<SCOPE_TYPE>:<SCOPE_ID>'.

    This function takes a string input representing a scope and deserializes it into `Scope` object
    containing the scope type, scope ID (if applicable), and an optional additional scope ID.

    The input should adhere to one of the following formats:
    1. '<SCOPE_TYPE>' (for system scope)
    2. '<SCOPE_TYPE>:<SCOPE_ID>'

    Scope types and their corresponding ID formats:
    - system: No ID (covers the whole system)
    - domain: String ID
    - project: UUID
    - user: UUID

    Args:
        value (str): The input string to deserialize, in one of the specified formats.

    Returns:
        One of [SystemScope, DomainScope, ProjectScope, UserScope] object.

    Raises:
        rbac.exceptions.InvalidScope:
            If the input string does not conform to the expected formats or if the scope type is invalid.
        ValueError:
            If the scope ID format doesn't match the expected type for the given scope.

    Examples:
        >>> deserialize_scope("system")
        SystemScope()
        >>> deserialize_scope("domain:default")
        DomainScope("default")
        >>> deserialize_scope("project:123e4567-e89b-12d3-a456-426614174000")
        ProjectScope(UUID('123e4567-e89b-12d3-a456-426614174000'))
        >>> deserialize_scope("user:123e4567-e89b-12d3-a456-426614174000")
        UserScope(UUID('123e4567-e89b-12d3-a456-426614174000'))
    """
    for scope in (SystemScope, DomainScope, ProjectScope, UserScope):
        try:
            return scope.deserialize(val)
        except ScopeTypeMismatch:
            continue
    else:
        raise InvalidScope(f"Invalid scope (s: {scope})")


# Extra scope is to address some scopes that contain specific object types
# such as registries for images, scaling groups for agents, storage hosts for vfolders etc.
class ExtraScope:
    pass


@dataclass(frozen=True)
class StorageHost(ExtraScope):
    name: str


@dataclass(frozen=True)
class ImageRegistry(ExtraScope):
    name: str


@dataclass(frozen=True)
class ScalingGroup(ExtraScope):
    name: str


@dataclass(frozen=True)
class ContainerRegistryScope(ExtraScope):
    registry_id: uuid.UUID

    def __str__(self) -> str:
        return f"container_registry:{str(self.registry_id)}"

    def __repr__(self) -> str:
        return self.__str__()

    @classmethod
    def parse(cls, raw: str) -> ContainerRegistryScope:
        _scope_type, _, registry_id = raw.partition(":")
        return cls(uuid.UUID(registry_id))


ObjectType = TypeVar("ObjectType")
ObjectIDType = TypeVar("ObjectIDType")


@dataclass
class AbstractPermissionContext(
    Generic[PermissionType, ObjectType, ObjectIDType], metaclass=ABCMeta
):
    """
    Define permissions under given User, Project or Domain scopes.
    Each field of this class represents a mapping of ["accessible scope id", "permissions under the scope"].
    For example, `project` field has a mapping of ["accessible project id", "permissions under the project"].
    {
        "PROJECT_A_ID": {"READ", "WRITE", "DELETE"}
        "PROJECT_B_ID": {"READ"}
    }

    `additional` and `overriding` fields have a mapping of ["object id", "permissions applied to the object"].
    `additional` field is used to add permissions to specific objects. It can be used for admins.
    `overriding` field is used to address exceptional cases such as permission overriding or cover other scopes(scaling groups or storage hosts etc).
    """

    user_id_to_permission_map: Mapping[uuid.UUID, frozenset[PermissionType]] = field(
        default_factory=dict
    )
    project_id_to_permission_map: Mapping[uuid.UUID, frozenset[PermissionType]] = field(
        default_factory=dict
    )
    domain_name_to_permission_map: Mapping[str, frozenset[PermissionType]] = field(
        default_factory=dict
    )

    object_id_to_additional_permission_map: Mapping[ObjectIDType, frozenset[PermissionType]] = (
        field(default_factory=dict)
    )
    object_id_to_overriding_permission_map: Mapping[ObjectIDType, frozenset[PermissionType]] = (
        field(default_factory=dict)
    )

    def filter_by_permission(self, permission_to_include: PermissionType) -> None:
        self.user_id_to_permission_map = {
            uid: permissions
            for uid, permissions in self.user_id_to_permission_map.items()
            if permission_to_include in permissions
        }
        self.project_id_to_permission_map = {
            pid: permissions
            for pid, permissions in self.project_id_to_permission_map.items()
            if permission_to_include in permissions
        }
        self.domain_name_to_permission_map = {
            dname: permissions
            for dname, permissions in self.domain_name_to_permission_map.items()
            if permission_to_include in permissions
        }
        self.object_id_to_additional_permission_map = {
            obj_id: permissions
            for obj_id, permissions in self.object_id_to_additional_permission_map.items()
            if permission_to_include in permissions
        }
        self.object_id_to_overriding_permission_map = {
            obj_id: permissions
            for obj_id, permissions in self.object_id_to_overriding_permission_map.items()
            if permission_to_include in permissions
        }

    def merge(self, trgt: Self) -> None:
        def _merge_map(
            src: Mapping[Any, frozenset[PermissionType]],
            trgt: Mapping[Any, frozenset[PermissionType]],
        ) -> dict[Any, frozenset[PermissionType]]:
            val = {}
            for key in {*src.keys(), *trgt.keys()}:
                val[key] = src.get(key, frozenset()) | trgt.get(key, frozenset())
            return val

        self.user_id_to_permission_map = _merge_map(
            self.user_id_to_permission_map, trgt.user_id_to_permission_map
        )
        self.project_id_to_permission_map = _merge_map(
            self.project_id_to_permission_map, trgt.project_id_to_permission_map
        )
        self.domain_name_to_permission_map = _merge_map(
            self.domain_name_to_permission_map, trgt.domain_name_to_permission_map
        )
        self.object_id_to_additional_permission_map = _merge_map(
            self.object_id_to_additional_permission_map, trgt.object_id_to_additional_permission_map
        )
        self.object_id_to_overriding_permission_map = _merge_map(
            self.object_id_to_overriding_permission_map, trgt.object_id_to_overriding_permission_map
        )

    def __add__(self, other: Self) -> Self:
        self.merge(other)
        return self

    @abstractmethod
    async def build_query(self) -> sa.sql.Select | None:
        pass

    @abstractmethod
    async def calculate_final_permission(self, rbac_obj: ObjectType) -> frozenset[PermissionType]:
        """
        Calculate the final permissions applied to the given object based on the fields in this class.
        """
        pass


PermissionContextType = TypeVar("PermissionContextType", bound=AbstractPermissionContext)


class AbstractPermissionContextBuilder(
    Generic[PermissionType, PermissionContextType], metaclass=ABCMeta
):
    @abstractmethod
    async def calculate_permission(
        self,
        ctx: ClientContext,
        target_scope: ScopeType,
    ) -> frozenset[PermissionType]:
        pass

    @classmethod
    async def _calculate_permission_by_predefined_roles(
        cls,
        roles: Iterable[PredefinedRole],
    ) -> frozenset[PermissionType]:
        result: frozenset[PermissionType] = frozenset()
        for role in roles:
            result |= await cls._calculate_permission_by_role(role)
        return result

    @classmethod
    async def _calculate_permission_by_role(
        cls,
        role: PredefinedRole,
    ) -> frozenset[PermissionType]:
        # This forces to implement a get_permission() method for each Built-in role.
        match role:
            case PredefinedRole.OWNER:
                return await cls._permission_for_owner()
            case PredefinedRole.ADMIN:
                return await cls._permission_for_admin()
            case PredefinedRole.MONITOR:
                return await cls._permission_for_monitor()
            case PredefinedRole.PRIVILEGED_MEMBER:
                return await cls._permission_for_privileged_member()
            case PredefinedRole.MEMBER:
                return await cls._permission_for_member()

    async def build(
        self,
        ctx: ClientContext,
        target_scope: ScopeType,
        requested_permission: PermissionType,
    ) -> PermissionContextType:
        match target_scope:
            case SystemScope():
                permission_ctx = await self.build_ctx_in_system_scope(ctx)
            case DomainScope():
                permission_ctx = await self.build_ctx_in_domain_scope(ctx, target_scope)
            case ProjectScope():
                permission_ctx = await self.build_ctx_in_project_scope(ctx, target_scope)
            case UserScope():
                permission_ctx = await self.build_ctx_in_user_scope(ctx, target_scope)
        permission_ctx.filter_by_permission(requested_permission)
        return permission_ctx

    @abstractmethod
    async def build_ctx_in_system_scope(self, ctx: ClientContext) -> PermissionContextType:
        pass

    @abstractmethod
    async def build_ctx_in_domain_scope(
        self,
        ctx: ClientContext,
        scope: DomainScope,
    ) -> PermissionContextType:
        pass

    @abstractmethod
    async def build_ctx_in_project_scope(
        self, ctx: ClientContext, scope: ProjectScope
    ) -> PermissionContextType:
        pass

    @abstractmethod
    async def build_ctx_in_user_scope(
        self, ctx: ClientContext, scope: UserScope
    ) -> PermissionContextType:
        pass

    @classmethod
    @abstractmethod
    async def _permission_for_owner(
        cls,
    ) -> frozenset[PermissionType]:
        pass

    @classmethod
    @abstractmethod
    async def _permission_for_admin(
        cls,
    ) -> frozenset[PermissionType]:
        pass

    @classmethod
    @abstractmethod
    async def _permission_for_monitor(
        cls,
    ) -> frozenset[PermissionType]:
        pass

    @classmethod
    @abstractmethod
    async def _permission_for_privileged_member(
        cls,
    ) -> frozenset[PermissionType]:
        pass

    @classmethod
    @abstractmethod
    async def _permission_for_member(
        cls,
    ) -> frozenset[PermissionType]:
        pass


class RBACModel(Generic[PermissionType]):
    @property
    @abstractmethod
    def permissions(self) -> Container[PermissionType]:
        pass


T_RBACModel = TypeVar("T_RBACModel", bound=RBACModel)
T_PropertyReturn = TypeVar("T_PropertyReturn")


def required_permission(permission: PermissionType):
    def wrapper(
        property_func: Callable[[T_RBACModel], T_PropertyReturn],
    ) -> Callable[[T_RBACModel], Optional[T_PropertyReturn]]:
        def wrapped(self: T_RBACModel) -> Optional[T_PropertyReturn]:
            if permission in self.permissions:
                return property_func(self)
            else:
                return None

        return wrapped

    return wrapper



from dataclasses import dataclass
from typing import Self

from ai.backend.manager.repositories.agent.repository import AgentRepository
from ai.backend.manager.repositories.types import RepositoryArgs


@dataclass
class AgentRepositories:
    repository: AgentRepository

    @classmethod
    def create(cls, args: RepositoryArgs) -> Self:
        repository = AgentRepository(
            args.db,
            args.valkey_image_client,
            args.valkey_live_client,
            args.config_provider,
        )

        return cls(
            repository=repository,
        )



from __future__ import annotations

import asyncio
import logging
import uuid
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterator,
    Final,
    Iterable,
    Mapping,
    Optional,
    Tuple,
)
from weakref import WeakSet

import aiohttp_cors
import attrs
import sqlalchemy as sa
import trafaret as t
from aiohttp import web
from aiohttp_sse import sse_response

from ai.backend.common import validators as tx
from ai.backend.common.events.dispatcher import (
    EventDispatcher,
)
from ai.backend.common.events.event_types.kernel.broadcast import (
    BaseKernelEvent,
    KernelCancelledBroadcastEvent,
    KernelCreatingBroadcastEvent,
    KernelPreparingBroadcastEvent,
    KernelPullingBroadcastEvent,
    KernelStartedBroadcastEvent,
    KernelTerminatedBroadcastEvent,
    KernelTerminatingBroadcastEvent,
)
from ai.backend.common.events.event_types.session.broadcast import (
    BaseSessionEvent,
    SchedulingBroadcastEvent,
    SessionCancelledBroadcastEvent,
    SessionEnqueuedBroadcastEvent,
    SessionFailureBroadcastEvent,
    SessionSuccessBroadcastEvent,
    SessionTerminatedBroadcastEvent,
    SessionTerminatingBroadcastEvent,
)
from ai.backend.common.events.hub import WILDCARD
from ai.backend.common.events.hub.propagators.cache import WithCachePropagator
from ai.backend.common.events.types import EventCacheDomain, EventDomain
from ai.backend.common.json import dump_json_str
from ai.backend.common.types import AgentId
from ai.backend.logging import BraceStyleAdapter

from ..errors.common import GenericForbidden
from ..errors.kernel import SessionNotFound
from ..errors.resource import ProjectNotFound
from ..events.hub.propagators.session import SessionEventPropagator
from ..exceptions import InvalidArgument
from ..models import UserRole, groups
from ..models.session import SessionRow
from ..types import Sentinel
from .auth import auth_required
from .manager import READ_ALLOWED, server_status_required
from .utils import check_api_params

if TYPE_CHECKING:
    from .context import RootContext
    from .types import CORSOptions, WebMiddleware

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

sentinel: Final = Sentinel.TOKEN

SessionEventInfo = Tuple[str, dict, str, Optional[int]]


@server_status_required(READ_ALLOWED)
@auth_required
@check_api_params(
    t.Dict({
        tx.AliasedKey(["name", "sessionName"], default="*") >> "session_name": t.String,
        t.Key("ownerAccessKey", default=None) >> "owner_access_key": t.Null | t.String,
        t.Key("sessionId", default=None) >> "session_id": t.Null | tx.UUID,
        # NOTE: if set, sessionId overrides sessionName and ownerAccessKey parameters.
        tx.AliasedKey(["group", "groupName"], default="*") >> "group_name": t.String,
        t.Key("scope", default="*"): t.String,
    })
)
async def push_session_events(
    request: web.Request,
    params: Mapping[str, Any],
) -> web.StreamResponse:
    root_ctx: RootContext = request.app["_root.context"]
    session_name = params["session_name"]
    session_id = params["session_id"]
    scope = params["scope"]
    user_role = request["user"]["role"]
    user_uuid = request["user"]["uuid"]
    access_key = params["owner_access_key"]
    if access_key is None:
        access_key = request["keypair"]["access_key"]
    if user_role == UserRole.USER:
        if access_key != request["keypair"]["access_key"]:
            raise GenericForbidden
    group_name = params["group_name"]
    if scope == "*":
        # Coalesce the legacy default value.
        scope = "session,kernel"

    log.info(
        "PUSH_SESSION_EVENTS (ak:{}, s:{} / s:{}, g:{}, scope:{})",
        access_key,
        session_name,
        session_id,
        group_name,
        scope,
    )
    priv_ctx: PrivateContext = request.app["events.context"]
    current_task = asyncio.current_task()
    assert current_task is not None
    priv_ctx.active_tasks.add(current_task)

    # Resolve session name to session ID
    if session_name == "*":
        session_id = WILDCARD
    else:
        async with root_ctx.db.begin_readonly_session(isolation_level="READ COMMITTED") as db_sess:
            rows = await SessionRow.match_sessions(
                db_sess, session_name, access_key, allow_prefix=False
            )
            if not rows:
                raise SessionNotFound
            session_id = rows[0].id

    # Resolve group name to group ID
    if group_name == "*":
        group_id = WILDCARD
    else:
        async with root_ctx.db.begin_readonly(isolation_level="READ COMMITTED") as conn:
            query = sa.select([groups.c.id]).select_from(groups).where(groups.c.name == group_name)
            result = await conn.execute(query)
            row = result.first()
            if row is None:
                raise ProjectNotFound
            group_id = row["id"]

    filters = {
        "user_role": user_role,
        "user_uuid": user_uuid,
        "domain_name": request["user"]["domain_name"],
        "group_id": group_id,
        "session_name": session_name,
        "session_id": session_id,
        "access_key": access_key,
    }
    aliases = []
    for item in scope.split(","):
        match item:
            case "session":
                aliases.append((EventDomain.SESSION, str(session_id)))
            case "kernel":
                aliases.append((EventDomain.KERNEL, str(session_id)))
            case _:
                raise InvalidArgument(f"Invalid scope: {scope}")

    async with sse_response(request) as resp:
        while not resp.prepared:
            await asyncio.sleep(0.1)
        propagator = SessionEventPropagator(resp, root_ctx.db, filters)
        root_ctx.event_hub.register_event_propagator(propagator, aliases)
        try:
            await resp.wait()
        finally:
            root_ctx.event_hub.unregister_event_propagator(propagator.id())
            await propagator.close()
            return resp


@server_status_required(READ_ALLOWED)
@auth_required
@check_api_params(
    t.Dict({
        tx.AliasedKey(["task_id", "taskId"]): tx.UUID,
    })
)
async def push_background_task_events(
    request: web.Request,
    params: Mapping[str, Any],
) -> web.StreamResponse:
    root_ctx: RootContext = request.app["_root.context"]
    priv_ctx: PrivateContext = request.app["events.context"]
    task_id: uuid.UUID = params["task_id"]
    access_key = request["keypair"]["access_key"]
    log.info("PUSH_BACKGROUND_TASK_EVENTS (ak:{}, t:{})", access_key, task_id)
    current_task = asyncio.current_task()
    assert current_task is not None
    priv_ctx.active_tasks.add(current_task)
    async with sse_response(request) as resp:
        propagator = WithCachePropagator(root_ctx.event_fetcher)
        root_ctx.event_hub.register_event_propagator(
            propagator, [(EventDomain.BGTASK, str(task_id))]
        )
        try:
            cache_id = EventCacheDomain.BGTASK.cache_id(str(task_id))
            async for event in propagator.receive(cache_id):
                user_event = event.user_event()
                if user_event is None:
                    log.warning(
                        "Received unsupported user event: {}",
                        event.event_name(),
                    )
                    continue
                await resp.send(
                    dump_json_str(user_event.user_event_mapping()),
                    event=user_event.event_name(),
                    retry=user_event.retry_count(),
                )
                if user_event.is_close_event():
                    log.debug(
                        "Received close event: {}",
                        user_event.event_name(),
                    )
                    break
            await resp.send(dump_json_str({}), event="server_close")
        finally:
            root_ctx.event_hub.unregister_event_propagator(propagator.id())
    return resp


async def _propagate_events(
    app: web.Application,
    agent_id: AgentId,
    event: BaseSessionEvent | BaseKernelEvent | SchedulingBroadcastEvent,
) -> None:
    """
    A private connector from EventDispatcher subscription to EventHub.
    """
    root_ctx: RootContext = app["_root.context"]
    log.trace("api.events._propagate_event({!r})", event)
    await root_ctx.event_hub.propagate_event(event)


@attrs.define(slots=True, auto_attribs=True)
class PrivateContext:
    active_tasks: WeakSet[asyncio.Task[web.StreamResponse]] = attrs.field(factory=WeakSet)


async def events_app_ctx(app: web.Application) -> AsyncIterator[None]:
    """
    Initialize events application context.
    Note: Event subscriptions are now handled by the event hub architecture,
    but we keep the legacy handlers for backward compatibility during transition.
    """
    root_ctx: RootContext = app["_root.context"]
    event_dispatcher: EventDispatcher = root_ctx.event_dispatcher

    # Keep legacy event dispatcher subscriptions for backward compatibility
    # These now delegate to the event hub
    event_dispatcher.subscribe(SessionEnqueuedBroadcastEvent, app, _propagate_events)
    # event_dispatcher.subscribe(SessionScheduledBroadcastEvent, app, _propagate_events)  # replaced by SchedulingBroadcastEvent
    event_dispatcher.subscribe(KernelPreparingBroadcastEvent, app, _propagate_events)
    event_dispatcher.subscribe(KernelPullingBroadcastEvent, app, _propagate_events)
    event_dispatcher.subscribe(KernelCreatingBroadcastEvent, app, _propagate_events)
    event_dispatcher.subscribe(KernelStartedBroadcastEvent, app, _propagate_events)
    # event_dispatcher.subscribe(SessionPreparingBroadcastEvent, app, _propagate_events)  # replaced by SchedulingBroadcastEvent
    # event_dispatcher.subscribe(SessionCreatingBroadcastEvent, app, _propagate_events)  # replaced by SchedulingBroadcastEvent
    # event_dispatcher.subscribe(SessionStartedBroadcastEvent, app, _propagate_events)  # replaced by SchedulingBroadcastEvent
    event_dispatcher.subscribe(KernelTerminatingBroadcastEvent, app, _propagate_events)
    event_dispatcher.subscribe(KernelTerminatedBroadcastEvent, app, _propagate_events)
    event_dispatcher.subscribe(KernelCancelledBroadcastEvent, app, _propagate_events)
    event_dispatcher.subscribe(SessionTerminatingBroadcastEvent, app, _propagate_events)
    event_dispatcher.subscribe(SessionTerminatedBroadcastEvent, app, _propagate_events)
    event_dispatcher.subscribe(SessionCancelledBroadcastEvent, app, _propagate_events)
    event_dispatcher.subscribe(SessionSuccessBroadcastEvent, app, _propagate_events)
    event_dispatcher.subscribe(SessionFailureBroadcastEvent, app, _propagate_events)

    # NOTE: SchedulingBroadcastEvent will replace most other session-related events
    #       since the Sokovan scheduler has been introduced.
    #       Currently its propagation to the Event Hub is already handled in the
    #       scheduler.dispatcher module.
    # FYI: In the future, we don't have to subscribe the events manually as the
    #      event dispatching mechanism will be centralized and migrated into the
    #      Event Hub's implementation detail.
    yield


async def events_shutdown(app: web.Application) -> None:
    """
    Shutdown handler for events app.
    Note: No longer needs to handle session event queues as they are managed by event hub.
    """
    root_ctx: RootContext = app["_root.context"]
    priv_ctx: PrivateContext = app["events.context"]
    await root_ctx.event_hub.shutdown()
    cancelled_tasks = []
    for task in priv_ctx.active_tasks:
        if not task.done():
            task.cancel()
            cancelled_tasks.append(task)
    await asyncio.gather(*cancelled_tasks, return_exceptions=True)


def create_app(
    default_cors_options: CORSOptions,
) -> Tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app["prefix"] = "events"
    app["events.context"] = PrivateContext()
    app["api_versions"] = (3, 4)
    app.on_shutdown.append(events_shutdown)
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    add_route = app.router.add_route
    app.cleanup_ctx.append(events_app_ctx)
    cors.add(add_route("GET", r"/background-task", push_background_task_events))
    cors.add(add_route("GET", r"/session", push_session_events))
    return app, []



"""Base classes for session preparation rules."""

from abc import ABC, abstractmethod
from typing import Any

from ai.backend.manager.repositories.scheduler.types.session_creation import (
    SessionCreationContext,
    SessionCreationSpec,
)


class SessionPreparerRule(ABC):
    """
    Abstract base class for session preparer rules.
    Each rule implements specific preparation logic for session creation.
    """

    @abstractmethod
    def name(self) -> str:
        """Return the preparer rule name."""
        raise NotImplementedError

    @abstractmethod
    def prepare(
        self,
        spec: SessionCreationSpec,
        context: SessionCreationContext,
        preparation_data: dict[str, Any],
    ) -> None:
        """
        Prepare/transform data for session creation.
        Updates the preparation_data dict in-place.

        Args:
            spec: Session creation specification
            context: Pre-fetched context with all required data
            preparation_data: Dictionary to store prepared data (modified in-place)
        """
        raise NotImplementedError



"""empty message

Revision ID: ff4bfca66bf8
Revises: 0e558d06e0e3, 352fa4f88f61
Create Date: 2018-12-24 22:42:54.188099

"""

# revision identifiers, used by Alembic.
revision = "ff4bfca66bf8"
down_revision = ("0e558d06e0e3", "352fa4f88f61")
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass



import logging
from datetime import datetime, timedelta
from decimal import Decimal
from typing import Any, Mapping, Optional, cast
from uuid import UUID

import msgpack
import sqlalchemy as sa
from dateutil.tz import tzutc
from sqlalchemy.ext.asyncio import AsyncSession as SASession
from sqlalchemy.orm import joinedload, load_only, noload

from ai.backend.common.clients.valkey_client.valkey_stat.client import ValkeyStatClient
from ai.backend.common.metrics.metric import LayerType
from ai.backend.common.utils import nmget
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.data.keypair.types import KeyPairCreator
from ai.backend.manager.data.permission.id import ObjectId, ScopeId
from ai.backend.manager.data.permission.types import EntityType, ScopeType
from ai.backend.manager.data.user.types import UserCreateResultData, UserCreator, UserData
from ai.backend.manager.decorators.repository_decorator import (
    create_layer_aware_repository_decorator,
)
from ai.backend.manager.defs import DEFAULT_KEYPAIR_RATE_LIMIT, DEFAULT_KEYPAIR_RESOURCE_POLICY_NAME
from ai.backend.manager.errors.user import (
    KeyPairForbidden,
    KeyPairNotFound,
    UserConflict,
    UserCreationBadRequest,
    UserCreationFailure,
    UserModificationFailure,
    UserNotFound,
)
from ai.backend.manager.models import kernels
from ai.backend.manager.models.domain import DomainRow
from ai.backend.manager.models.group import ProjectType, association_groups_users, groups
from ai.backend.manager.models.kernel import RESOURCE_USAGE_KERNEL_STATUSES
from ai.backend.manager.models.keypair import KeyPairRow, generate_keypair_data, keypairs
from ai.backend.manager.models.user import UserRole, UserRow, UserStatus, users
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine
from ai.backend.manager.services.user.actions.modify_user import UserModifier

from ..permission_controller.role_manager import RoleManager

# Layer-specific decorator for user repository
repository_decorator = create_layer_aware_repository_decorator(LayerType.USER)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class UserRepository:
    _db: ExtendedAsyncSAEngine
    _role_manager: RoleManager

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db = db
        self._role_manager = RoleManager()

    @repository_decorator()
    async def get_user_by_uuid(self, user_uuid: UUID) -> UserData:
        """
        Get user by UUID without ownership validation.
        Admin-only operation.
        """
        async with self._db.begin_readonly_session() as db_session:
            user_row = await self._get_user_by_uuid(db_session, user_uuid)
            return user_row.to_data()

    @repository_decorator()
    async def get_by_email_validated(
        self,
        email: str,
    ) -> UserData:
        """
        Get user by email with ownership validation.
        Returns None if user not found or access denied.
        """
        async with self._db.begin_session() as session:
            user_row = await self._get_user_by_email(session, email)
            return UserData.from_row(user_row)

    @repository_decorator()
    async def create_user_validated(
        self, user_creator: UserCreator, group_ids: Optional[list[str]]
    ) -> UserCreateResultData:
        """
        Create a new user with default keypair and group associations.
        """
        domain_name = user_creator.domain_name
        email = user_creator.email
        user_name = user_creator.username

        async with self._db.begin_session() as db_session:
            # Check if domain exists before creating user
            domain_exists = await self._check_domain_exists(db_session, domain_name)
            if not domain_exists:
                raise UserCreationBadRequest(f"Domain '{domain_name}' does not exist.")

            # Check if user with the same email or username already exists
            duplicate_exists = await self._check_user_exists_with_email_or_username(
                db_session, email=email, username=user_name
            )
            if duplicate_exists:
                raise UserConflict(
                    f"User with email {email} or username {user_name} already exists."
                )
            try:
                # Insert user
                row = UserRow.from_creator(user_creator)
                db_session.add(row)
                await db_session.flush()
                await db_session.refresh(row)
            except sa.exc.IntegrityError as e:
                error_msg = str(e)
                raise UserCreationBadRequest(
                    f"Failed to create user due to database constraint violation: {error_msg}"
                ) from e

            if not row:
                raise UserCreationFailure("Failed to create user")
            created_user = row.to_data()

            # Create default keypair
            email = created_user.email
            keypair_creator = KeyPairCreator(
                is_active=(created_user.status == UserStatus.ACTIVE),
                is_admin=created_user.role in ["superadmin", "admin"],
                resource_policy=DEFAULT_KEYPAIR_RESOURCE_POLICY_NAME,
                rate_limit=DEFAULT_KEYPAIR_RATE_LIMIT,
            )
            generated = generate_keypair_data()
            kp_row = KeyPairRow.from_creator(keypair_creator, generated, created_user.id, email)
            db_session.add(kp_row)
            await db_session.flush()
            await db_session.refresh(kp_row)
            kp_data = kp_row.to_data()

            # Update user main_access_key
            row.main_access_key = kp_data.access_key
            created_user.main_access_key = kp_data.access_key

            # Add user to groups including model store project
            await self._add_user_to_groups(
                db_session, created_user.uuid, created_user.domain_name, group_ids or []
            )

            role = await self._role_manager.create_system_role(db_session, created_user)
            await self._role_manager.map_user_to_role(db_session, created_user.uuid, role.id)
            await self._role_manager.map_entity_to_scope(
                db_session,
                ObjectId(EntityType.USER, str(created_user.uuid)),
                ScopeId(ScopeType.DOMAIN, str(created_user.domain_name)),
            )

        return UserCreateResultData(created_user, kp_data)

    @repository_decorator()
    async def update_user_validated(
        self,
        email: str,
        modifier: UserModifier,
        requester_uuid: Optional[UUID],
    ) -> UserData:
        """
        Update user with ownership validation and handle role/group changes.
        """
        to_update = modifier.fields_to_update()
        async with self._db.begin() as conn:
            # Get current user data for validation
            current_user = await self._get_user_by_email_with_conn(conn, email)

            # Handle main_access_key validation
            main_access_key = modifier.main_access_key.optional_value()
            if main_access_key:
                await self._validate_and_update_main_access_key(conn, email, main_access_key)

            # Update user
            if modifier.password.optional_value():
                to_update["password_changed_at"] = sa.func.now()
            status = modifier.status.optional_value()
            if status is not None and status != current_user.status:
                to_update["status_info"] = "admin-requested"
            update_query = (
                sa.update(users).where(users.c.email == email).values(to_update).returning(users)
            )
            result = await conn.execute(update_query)
            updated_user = result.first()
            if not updated_user:
                raise UserModificationFailure("Failed to update user")

            # Handle role changes
            prev_role = current_user.role
            role = modifier.role.optional_value()
            if role is not None and role != prev_role:
                await self._sync_keypair_roles(conn, updated_user.uuid, role)

            # Handle group updates
            group_ids = modifier.group_ids_value
            if prev_role != updated_user.role and group_ids is None:
                await self._clear_user_groups(conn, updated_user.uuid)

            if group_ids is not None:
                await self._update_user_groups(
                    conn, updated_user.uuid, updated_user.domain_name, group_ids
                )
            res = UserData.from_row(updated_user)
        return res

    @repository_decorator()
    async def soft_delete_user_validated(self, email: str, requester_uuid: Optional[UUID]) -> None:
        """
        Soft delete user by setting status to DELETED and deactivating keypairs.
        """
        async with self._db.begin() as conn:
            # Deactivate all user keypairs
            await conn.execute(
                sa.update(keypairs).values(is_active=False).where(keypairs.c.user_id == email)
            )
            # Soft delete user
            await conn.execute(
                sa.update(users)
                .values(status=UserStatus.DELETED, status_info="admin-requested")
                .where(users.c.email == email)
            )

    async def _check_domain_exists(self, session: SASession, domain_name: str) -> bool:
        query = sa.select(DomainRow.name).where(DomainRow.name == domain_name)
        result = await session.scalar(query)
        result = cast(Optional[str], result)
        return result is not None

    async def _check_user_exists_with_email_or_username(
        self, session: SASession, *, email: str, username: str
    ) -> bool:
        query = sa.select(UserRow.uuid).where(
            sa.or_(UserRow.email == email, UserRow.username == username)
        )
        result = await session.scalar(query)
        result = cast(Optional[UUID], result)
        return result is not None

    async def _get_user_by_email(self, session: SASession, email: str) -> UserRow:
        """Private method to get user by email."""
        res = await session.scalar(sa.select(UserRow).where(UserRow.email == email))
        if res is None:
            raise UserNotFound(f"User with email {email} not found.")
        return res

    async def _get_user_by_uuid(self, session: SASession, user_uuid: UUID) -> UserRow:
        """Private method to get user by UUID."""
        res = await session.scalar(sa.select(UserRow).where(UserRow.uuid == user_uuid))
        if res is None:
            raise UserNotFound(f"User with UUID {user_uuid} not found.")
        return res

    async def _get_user_by_email_with_conn(self, conn, email: str) -> UserRow:
        """Private method to get user by email using connection."""
        result = await conn.execute(sa.select(users).where(users.c.email == email))
        res = result.first()
        if res is None:
            raise UserNotFound(f"User with email {email} not found.")
        return res

    def _validate_user_access(self, user_row: UserRow, requester_uuid: Optional[UUID]) -> bool:
        """Private method to validate user access - can be extended for ownership logic."""
        # For now, allow access - this can be extended with ownership validation
        return True

    async def _add_user_to_groups(
        self, db_session: SASession, user_uuid: UUID, domain_name: str, group_ids: list[str]
    ) -> None:
        """Private method to add user to groups including model store project."""
        # Check for model store project
        model_store_query = sa.select(groups.c.id).where(groups.c.type == ProjectType.MODEL_STORE)
        model_store_project = (await db_session.execute(model_store_query)).first()

        gids_to_join = list(group_ids)
        if model_store_project:
            gids_to_join.append(model_store_project["id"])

        if gids_to_join:
            query = (
                sa.select(groups.c.id)
                .select_from(groups)
                .where(groups.c.domain_name == domain_name)
                .where(groups.c.id.in_(gids_to_join))
            )
            grps = (await db_session.execute(query)).all()
            if grps:
                group_data = [{"user_id": user_uuid, "group_id": grp.id} for grp in grps]
                group_insert_query = sa.insert(association_groups_users).values(group_data)
                await db_session.execute(group_insert_query)
            else:
                log.warning(
                    "No valid groups found to add user {0} in domain {1}", user_uuid, domain_name
                )
        else:
            log.info("Adding new user {0} with no groups in domain {1}", user_uuid, domain_name)

    async def _validate_and_update_main_access_key(
        self, conn, email: str, main_access_key: str
    ) -> None:
        """Private method to validate and update main access key."""
        session = SASession(conn)
        keypair_query = (
            sa.select(KeyPairRow)
            .where(KeyPairRow.access_key == main_access_key)
            .options(
                noload("*"),
                joinedload(KeyPairRow.user_row).options(load_only(UserRow.email)),
            )
        )
        keypair_row = (await session.scalars(keypair_query)).first()
        if not keypair_row:
            raise KeyPairNotFound("Cannot set non-existing access key as the main access key.")
        if keypair_row.user_row.email != email:
            raise KeyPairForbidden("Cannot set another user's access key as the main access key.")

        await conn.execute(
            sa.update(users).where(users.c.email == email).values(main_access_key=main_access_key)
        )

    async def _sync_keypair_roles(self, conn, user_uuid: UUID, new_role: UserRole) -> None:
        """Private method to sync keypair roles with user role."""
        from sqlalchemy.sql.expression import bindparam

        result = await conn.execute(
            sa.select([
                keypairs.c.user,
                keypairs.c.is_active,
                keypairs.c.is_admin,
                keypairs.c.access_key,
            ])
            .select_from(keypairs)
            .where(keypairs.c.user == user_uuid)
            .order_by(sa.desc(keypairs.c.is_admin))
            .order_by(sa.desc(keypairs.c.is_active))
        )

        if new_role in [UserRole.SUPERADMIN, UserRole.ADMIN]:
            # User becomes admin - set first keypair as active admin
            kp = result.first()
            kp_data = {}
            if not kp.is_admin:
                kp_data["is_admin"] = True
            if not kp.is_active:
                kp_data["is_active"] = True
            if kp_data:
                await conn.execute(
                    sa.update(keypairs).values(kp_data).where(keypairs.c.user == user_uuid)
                )
        else:
            # User becomes non-admin - update keypairs accordingly
            rows = result.fetchall()
            kp_updates = []
            for idx, row in enumerate(rows):
                kp_data = {
                    "b_access_key": row.access_key,
                    "is_admin": row.is_admin,
                    "is_active": row.is_active,
                }
                if idx == 0:
                    kp_data["is_admin"] = False
                    kp_updates.append(kp_data)
                    continue
                if row.is_admin and row.is_active:
                    kp_data["is_active"] = False
                    kp_updates.append(kp_data)

            if kp_updates:
                await conn.execute(
                    sa.update(keypairs)
                    .values({
                        "is_admin": bindparam("is_admin"),
                        "is_active": bindparam("is_active"),
                    })
                    .where(keypairs.c.access_key == bindparam("b_access_key")),
                    kp_updates,
                )

    async def _clear_user_groups(self, conn, user_uuid: UUID) -> None:
        """Private method to clear user's group associations."""
        await conn.execute(
            sa.delete(association_groups_users).where(
                association_groups_users.c.user_id == user_uuid
            )
        )

    async def _update_user_groups(
        self, conn, user_uuid: UUID, domain_name: str, group_ids: list[str]
    ) -> None:
        """Private method to update user's group associations."""
        # Clear existing groups
        await self._clear_user_groups(conn, user_uuid)

        # Add to new groups
        result = await conn.execute(
            sa.select([groups.c.id])
            .select_from(groups)
            .where(groups.c.domain_name == domain_name)
            .where(groups.c.id.in_(group_ids))
        )
        grps = result.fetchall()
        if grps:
            values = [{"user_id": user_uuid, "group_id": grp.id} for grp in grps]
            await conn.execute(sa.insert(association_groups_users).values(values))

    @repository_decorator()
    async def get_user_time_binned_monthly_stats(
        self,
        user_uuid: UUID,
        valkey_stat_client: ValkeyStatClient,
    ) -> list[dict[str, Any]]:
        """
        Generate time-binned (15 min) stats for the last one month for a specific user.
        """
        return await self._get_time_binned_monthly_stats(user_uuid, valkey_stat_client)

    async def _get_time_binned_monthly_stats(
        self,
        user_uuid: Optional[UUID],
        valkey_stat_client: ValkeyStatClient,
    ) -> list[dict[str, Any]]:
        """
        Generate time-binned (15 min) stats for the last one month.
        """
        # Get all or user kernels for the last month from DB.
        time_window = 900  # 15 min
        stat_length = 2880  # 15 * 4 * 24 * 30
        now = datetime.now(tzutc())
        start_date = now - timedelta(days=30)

        async with self._db.begin_readonly() as conn:
            query = (
                sa.select([
                    kernels.c.id,
                    kernels.c.created_at,
                    kernels.c.terminated_at,
                    kernels.c.occupied_slots,
                ])
                .select_from(kernels)
                .where(
                    (kernels.c.terminated_at >= start_date)
                    & (kernels.c.status.in_(RESOURCE_USAGE_KERNEL_STATUSES)),
                )
                .order_by(sa.asc(kernels.c.created_at))
            )
            if user_uuid is not None:
                query = query.where(kernels.c.user_uuid == user_uuid)
            result = await conn.execute(query)
            rows = result.fetchall()

        # Build time-series of time-binned stats.
        start_date_ts = start_date.timestamp()
        time_series_list: list[dict[str, Any]] = [
            {
                "date": start_date_ts + (idx * time_window),
                "num_sessions": {
                    "value": 0,
                    "unit_hint": "count",
                },
                "cpu_allocated": {
                    "value": 0,
                    "unit_hint": "count",
                },
                "mem_allocated": {
                    "value": 0,
                    "unit_hint": "bytes",
                },
                "gpu_allocated": {
                    "value": 0,
                    "unit_hint": "count",
                },
                "io_read_bytes": {
                    "value": 0,
                    "unit_hint": "bytes",
                },
                "io_write_bytes": {
                    "value": 0,
                    "unit_hint": "bytes",
                },
                "disk_used": {
                    "value": 0,
                    "unit_hint": "bytes",
                },
            }
            for idx in range(stat_length)
        ]

        kernel_ids = [str(row["id"]) for row in rows]
        raw_stats = await valkey_stat_client.get_user_kernel_statistics_batch(kernel_ids)

        for row, raw_stat in zip(rows, raw_stats):
            if raw_stat is not None:
                last_stat = msgpack.unpackb(raw_stat)
                io_read_byte = int(nmget(last_stat, "io_read.current", 0))
                io_write_byte = int(nmget(last_stat, "io_write.current", 0))
                disk_used = int(nmget(last_stat, "io_scratch_size.stats.max", 0, "/"))
            else:
                io_read_byte = 0
                io_write_byte = 0
                disk_used = 0

            occupied_slots: Mapping[str, Any] = row.occupied_slots
            kernel_created_at: float = row.created_at.timestamp()
            kernel_terminated_at: float = row.terminated_at.timestamp()
            cpu_value = int(occupied_slots.get("cpu", 0))
            mem_value = int(occupied_slots.get("mem", 0))
            cuda_device_value = int(occupied_slots.get("cuda.devices", 0))
            cuda_share_value = Decimal(occupied_slots.get("cuda.shares", 0))

            start_index = int((kernel_created_at - start_date_ts) // time_window)
            end_index = int((kernel_terminated_at - start_date_ts) // time_window) + 1
            if start_index < 0:
                start_index = 0
            for time_series in time_series_list[start_index:end_index]:
                time_series["num_sessions"]["value"] += 1
                time_series["cpu_allocated"]["value"] += cpu_value
                time_series["mem_allocated"]["value"] += mem_value
                time_series["gpu_allocated"]["value"] += cuda_device_value
                time_series["gpu_allocated"]["value"] += cuda_share_value
                time_series["io_read_bytes"]["value"] += io_read_byte
                time_series["io_write_bytes"]["value"] += io_write_byte
                time_series["disk_used"]["value"] += disk_used

        # Change Decimal type to float to serialize to JSON
        for time_series in time_series_list:
            time_series["gpu_allocated"]["value"] = float(time_series["gpu_allocated"]["value"])
        return time_series_list



"""Validator for user resource limits."""

from ai.backend.common.types import ResourceSlot
from ai.backend.manager.sokovan.scheduler.types import SessionWorkload, SystemSnapshot

from .exceptions import UserResourceQuotaExceeded
from .validator import ValidatorRule


class UserResourceLimitValidator(ValidatorRule):
    """
    Check if a session would exceed the user's resource quota.
    This corresponds to check_user_resource_limit predicate.
    """

    def name(self) -> str:
        """Return the validator name for predicates."""
        return "UserResourceLimitValidator"

    def success_message(self) -> str:
        """Return a message describing successful validation."""
        return "User has sufficient resource quota for the requested session"

    def validate(self, snapshot: SystemSnapshot, workload: SessionWorkload) -> None:
        # Get the user's resource policy
        policy = snapshot.resource_policy.user_policies.get(workload.user_uuid)
        if not policy:
            # If no user-specific policy, skip validation (no limits apply)
            return

        # Get current user occupancy
        user_occupied = snapshot.resource_occupancy.by_user.get(workload.user_uuid, ResourceSlot())

        # Check if adding this workload would exceed the limit
        total_after = user_occupied + workload.requested_slots
        if not (total_after <= policy.total_resource_slots):
            exceeded_msg = " ".join(f"{k}={v}" for k, v in policy.total_resource_slots.items() if v)
            raise UserResourceQuotaExceeded(
                f"Your main-keypair resource quota is exceeded. ({exceeded_msg})"
            )



"""Cluster configuration preparation rule."""

from typing import Any, cast

from ai.backend.common.types import KernelEnqueueingConfig
from ai.backend.manager.defs import DEFAULT_ROLE
from ai.backend.manager.repositories.scheduler.types.session_creation import (
    SessionCreationContext,
    SessionCreationSpec,
)

from .base import SessionPreparerRule


class ClusterConfigurationRule(SessionPreparerRule):
    """Prepares cluster kernel configurations."""

    def name(self) -> str:
        return "cluster_configuration"

    def prepare(
        self,
        spec: SessionCreationSpec,
        context: SessionCreationContext,
        preparation_data: dict[str, Any],
    ) -> None:
        """Prepare cluster kernel configurations."""
        if spec.cluster_size > 1:
            kernel_configs = self._prepare_cluster_kernels(spec)
        else:
            # Single kernel session
            kernel_configs = spec.kernel_specs.copy()
            if kernel_configs and "cluster_role" not in kernel_configs[0]:
                kernel_configs[0]["cluster_role"] = DEFAULT_ROLE

        preparation_data["kernel_configs"] = kernel_configs

    def _prepare_cluster_kernels(self, spec: SessionCreationSpec) -> list[KernelEnqueueingConfig]:
        """Prepare kernel configurations for multi-container cluster."""
        if len(spec.kernel_specs) == 1:
            # Replicate the single kernel spec for all cluster nodes
            main_kernel_config = spec.kernel_specs[0].copy()
            kernel_configs = []

            # Configure main kernel (cluster_idx = 1)
            main_kernel_config["cluster_role"] = "main"
            main_kernel_config["cluster_idx"] = 1
            main_kernel_config["local_rank"] = 0
            main_kernel_config["cluster_hostname"] = "main1"
            kernel_configs.append(main_kernel_config)

            # Configure sub kernels (cluster_idx starts from 1 for sub1, 2 for sub2, etc.)
            for i in range(spec.cluster_size - 1):
                sub_kernel_config = cast(KernelEnqueueingConfig, {**main_kernel_config})
                sub_kernel_config["cluster_role"] = "sub"
                sub_kernel_config["cluster_idx"] = i + 1  # sub1: 1, sub2: 2, ...
                sub_kernel_config["local_rank"] = i + 1  # sub1: 1, sub2: 2, ...
                sub_kernel_config["cluster_hostname"] = f"sub{i + 1}"
                kernel_configs.append(sub_kernel_config)
        else:
            # Use provided kernel specs - each container has its own config
            kernel_configs = []
            for idx, kernel_spec in enumerate(spec.kernel_specs):
                kernel_config = kernel_spec.copy()

                # Check if cluster_role is already defined
                if "cluster_role" not in kernel_config:
                    # Assign default roles based on position
                    if idx == 0:
                        kernel_config["cluster_role"] = DEFAULT_ROLE
                    else:
                        kernel_config["cluster_role"] = "sub"

                # Set cluster_idx if not already set
                if "cluster_idx" not in kernel_config:
                    if kernel_config["cluster_role"] == DEFAULT_ROLE:
                        kernel_config["cluster_idx"] = 1
                    else:
                        # For sub kernels, find the count of previous sub kernels
                        sub_count = sum(
                            1 for i in range(idx) if kernel_configs[i].get("cluster_role") == "sub"
                        )
                        kernel_config["cluster_idx"] = sub_count + 1

                # Set local_rank if not already set
                if "local_rank" not in kernel_config:
                    kernel_config["local_rank"] = idx

                # Set cluster_hostname if not already set
                if "cluster_hostname" not in kernel_config:
                    role = kernel_config["cluster_role"]
                    idx_val = kernel_config["cluster_idx"]
                    kernel_config["cluster_hostname"] = f"{role}{idx_val}"

                kernel_configs.append(kernel_config)

        return kernel_configs



"""
Hook for batch session type.
Triggers batch execution when the session transitions to running.
"""

import logging

from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.clients.agent.pool import AgentPool

from ..types import SessionTransitionData
from .base import AbstractSessionHook

log = BraceStyleAdapter(logging.getLogger(__name__))


class BatchSessionHook(AbstractSessionHook):
    _agent_pool: AgentPool

    def __init__(self, agent_pool: AgentPool) -> None:
        self._agent_pool = agent_pool

    async def on_transition_to_running(self, session: SessionTransitionData) -> None:
        main_kernel = session.main_kernel
        async with self._agent_pool._agent_cache.rpc_context(
            main_kernel.agent_id,
            invoke_timeout=30,
            order_key=main_kernel.kernel_id,
        ) as rpc:
            await rpc.call.trigger_batch_execution(
                str(session.session_id),
                str(main_kernel.kernel_id),
                main_kernel.startup_command or "",
                session.batch_timeout,
            )
        log.info(
            "Successfully triggered batch execution for session {} on agent {}",
            session.session_id,
            main_kernel.agent_id,
        )

    async def on_transition_to_terminated(self, session: SessionTransitionData) -> None:
        log.debug(
            "Batch session {} transitioning to TERMINATED",
            session.session_id,
        )



from typing import Optional

import sqlalchemy as sa

from ai.backend.common.metrics.metric import LayerType
from ai.backend.manager.data.container_registry.types import ContainerRegistryData
from ai.backend.manager.data.image.types import ImageStatus
from ai.backend.manager.decorators.repository_decorator import (
    create_layer_aware_repository_decorator,
)
from ai.backend.manager.errors.image import ContainerRegistryNotFound
from ai.backend.manager.models.container_registry import ContainerRegistryRow
from ai.backend.manager.models.image import ImageRow
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine, SASession

# Layer-specific decorator for container_registry repository
repository_decorator = create_layer_aware_repository_decorator(LayerType.CONTAINER_REGISTRY)


class ContainerRegistryRepository:
    _db: ExtendedAsyncSAEngine

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db = db

    @repository_decorator()
    async def get_by_registry_and_project(
        self,
        registry_name: str,
        project: Optional[str] = None,
    ) -> ContainerRegistryData:
        async with self._db.begin_readonly_session() as session:
            result = await self._get_by_registry_and_project(session, registry_name, project)
            if not result:
                raise ContainerRegistryNotFound()
            return result

    @repository_decorator()
    async def get_by_registry_name(self, registry_name: str) -> list[ContainerRegistryData]:
        async with self._db.begin_readonly_session() as session:
            return await self._get_by_registry_name(session, registry_name)

    @repository_decorator()
    async def get_all(self) -> list[ContainerRegistryData]:
        async with self._db.begin_readonly_session() as session:
            return await self._get_all(session)

    @repository_decorator()
    async def clear_images(
        self,
        registry_name: str,
        project: Optional[str] = None,
    ) -> ContainerRegistryData:
        async with self._db.begin_session() as session:
            # Clear images
            update_stmt = (
                sa.update(ImageRow)
                .where(ImageRow.registry == registry_name)
                .where(ImageRow.status != ImageStatus.DELETED)
                .values(status=ImageStatus.DELETED)
            )
            if project:
                update_stmt = update_stmt.where(ImageRow.project == project)

            await session.execute(update_stmt)

            # Return registry data
            result = await self._get_by_registry_and_project(session, registry_name, project)
            if not result:
                raise ContainerRegistryNotFound()
            return result

    @repository_decorator()
    async def get_known_registries(self) -> dict[str, str]:
        async with self._db.begin_readonly_session() as session:
            known_registries_map = await ContainerRegistryRow.get_known_container_registries(
                session
            )

            known_registries = {}
            for project, registries in known_registries_map.items():
                for registry_name, url in registries.items():
                    if project not in known_registries:
                        known_registries[f"{project}/{registry_name}"] = url.human_repr()

            return known_registries

    @repository_decorator()
    async def get_registry_row_for_scanner(
        self,
        registry_name: str,
        project: Optional[str] = None,
    ) -> ContainerRegistryRow:
        """
        Get the raw ContainerRegistryRow object needed for container registry scanner.
        Raises ContainerRegistryNotFound if registry is not found.
        """
        async with self._db.begin_readonly_session() as session:
            stmt = sa.select(ContainerRegistryRow).where(
                ContainerRegistryRow.registry_name == registry_name,
            )
            if project:
                stmt = stmt.where(ContainerRegistryRow.project == project)

            row: Optional[ContainerRegistryRow] = await session.scalar(stmt)
            if not row:
                raise ContainerRegistryNotFound()
            return row

    async def _get_by_registry_and_project(
        self,
        session: SASession,
        registry_name: str,
        project: Optional[str] = None,
    ) -> Optional[ContainerRegistryData]:
        stmt = sa.select(ContainerRegistryRow).where(
            ContainerRegistryRow.registry_name == registry_name,
        )
        if project:
            stmt = stmt.where(ContainerRegistryRow.project == project)

        row: Optional[ContainerRegistryRow] = await session.scalar(stmt)
        return row.to_dataclass() if row else None

    async def _get_by_registry_name(
        self,
        session: SASession,
        registry_name: str,
    ) -> list[ContainerRegistryData]:
        stmt = sa.select(ContainerRegistryRow).where(
            ContainerRegistryRow.registry_name == registry_name
        )
        result = await session.execute(stmt)
        rows: list[ContainerRegistryRow] = result.scalars().all()
        return [row.to_dataclass() for row in rows]

    async def _get_all(self, session: SASession) -> list[ContainerRegistryData]:
        stmt = sa.select(ContainerRegistryRow)
        result = await session.execute(stmt)
        rows: list[ContainerRegistryRow] = result.scalars().all()
        return [row.to_dataclass() for row in rows]



"""Handler for scheduling pending sessions."""

import logging
from typing import Optional

from ai.backend.common.events.dispatcher import EventProducer
from ai.backend.common.events.event_types.session.broadcast import (
    SchedulingBroadcastEvent,
)
from ai.backend.common.events.types import AbstractBroadcastEvent
from ai.backend.common.types import AccessKey
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.session.types import SessionStatus
from ai.backend.manager.defs import LockID
from ai.backend.manager.repositories.scheduler.repository import SchedulerRepository
from ai.backend.manager.scheduler.types import ScheduleType
from ai.backend.manager.sokovan.scheduler.results import ScheduleResult
from ai.backend.manager.sokovan.scheduler.scheduler import Scheduler
from ai.backend.manager.sokovan.scheduling_controller import SchedulingController

from .base import SchedulerHandler

log = BraceStyleAdapter(logging.getLogger(__name__))


class ScheduleSessionsHandler(SchedulerHandler):
    """Handler for scheduling pending sessions."""

    def __init__(
        self,
        scheduler: Scheduler,
        scheduling_controller: SchedulingController,
        event_producer: EventProducer,
        repository: SchedulerRepository,
    ):
        self._scheduler = scheduler
        self._scheduling_controller = scheduling_controller
        self._event_producer = event_producer
        self._repository = repository

    @classmethod
    def name(cls) -> str:
        """Get the name of the handler."""
        return "schedule-sessions"

    @property
    def lock_id(self) -> Optional[LockID]:
        """Lock for operations targeting PENDING sessions."""
        return LockID.LOCKID_SOKOVAN_TARGET_PENDING

    async def execute(self) -> ScheduleResult:
        """Schedule all pending sessions across scaling groups."""
        log.trace("Scheduling sessions across all scaling groups")
        return await self._scheduler.schedule_all_scaling_groups()

    async def post_process(self, result: ScheduleResult) -> None:
        """Request precondition check if sessions were scheduled and invalidate cache."""
        # Request next phase first
        await self._scheduling_controller.mark_scheduling_needed(ScheduleType.CHECK_PRECONDITION)
        log.info(
            "Scheduled {} sessions, requesting precondition check", len(result.scheduled_sessions)
        )

        # Invalidate cache for affected access keys
        affected_keys: set[AccessKey] = {
            session.access_key for session in result.scheduled_sessions
        }
        await self._repository.invalidate_kernel_related_cache(list(affected_keys))
        log.debug("Invalidated kernel-related cache for {} access keys", len(affected_keys))

        # Broadcast batch event for scheduled sessions
        events: list[AbstractBroadcastEvent] = [
            SchedulingBroadcastEvent(
                session_id=event_data.session_id,
                creation_id=event_data.creation_id,
                status_transition=str(SessionStatus.SCHEDULED),
                reason=event_data.reason,
            )
            for event_data in result.scheduled_sessions
        ]
        await self._event_producer.broadcast_events_batch(events)



"""Validator for pending session count limits."""

from ai.backend.manager.sokovan.scheduler.types import SessionWorkload, SystemSnapshot

from .exceptions import PendingSessionCountLimitExceeded
from .validator import ValidatorRule


class PendingSessionCountLimitValidator(ValidatorRule):
    """
    Check if creating a new pending session would exceed the allowed count.
    This corresponds to check_pending_session_count_limit predicate.
    """

    def name(self) -> str:
        """Return the validator name for predicates."""
        return "PendingSessionCountLimitValidator"

    def success_message(self) -> str:
        """Return a message describing successful validation."""
        return "Pending session count is within the allowed limit for the access key"

    def validate(self, snapshot: SystemSnapshot, workload: SessionWorkload) -> None:
        # Get the keypair's resource policy
        policy = snapshot.resource_policy.keypair_policies.get(workload.access_key)
        if not policy:
            # If no policy is defined, we can't validate - let it pass
            return

        # Check if there's a pending session count limit
        pending_count_limit = policy.max_pending_session_count
        if pending_count_limit is None:
            # No limit set, allow
            return

        # Get current pending sessions for this keypair
        pending_sessions = snapshot.pending_sessions.by_keypair.get(workload.access_key, [])
        current_pending_count = len(pending_sessions)

        # Check if creating this session would exceed the limit
        if current_pending_count >= pending_count_limit:
            raise PendingSessionCountLimitExceeded(
                f"You cannot create more than {pending_count_limit} pending session(s)."
            )



from typing import Optional

from ai.backend.manager.data.container_registry.types import ContainerRegistryData
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine

from .repository import ContainerRegistryRepository


class AdminContainerRegistryRepository:
    _repository: ContainerRegistryRepository

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._repository = ContainerRegistryRepository(db)

    async def clear_images_force(
        self,
        registry_name: str,
        project: Optional[str] = None,
    ) -> ContainerRegistryData:
        """
        Forcefully clear images from a container registry without any validation.
        This is an admin-only operation that should be used with caution.
        """
        return await self._repository.clear_images(registry_name, project)

    async def get_by_registry_and_project_force(
        self,
        registry_name: str,
        project: Optional[str] = None,
    ) -> ContainerRegistryData:
        """
        Forcefully get container registry by name and project without any validation.
        This is an admin-only operation that should be used with caution.
        """
        return await self._repository.get_by_registry_and_project(registry_name, project)

    async def get_by_registry_name_force(self, registry_name: str) -> list[ContainerRegistryData]:
        """
        Forcefully get container registries by name without any validation.
        This is an admin-only operation that should be used with caution.
        """
        return await self._repository.get_by_registry_name(registry_name)

    async def get_all_force(self) -> list[ContainerRegistryData]:
        """
        Forcefully get all container registries without any validation.
        This is an admin-only operation that should be used with caution.
        """
        return await self._repository.get_all()



from ai.backend.manager.actions.monitors.monitor import ActionMonitor
from ai.backend.manager.actions.processor import ActionProcessor
from ai.backend.manager.services.metric.actions.container import (
    ContainerMetricAction,
    ContainerMetricActionResult,
    ContainerMetricMetadataAction,
    ContainerMetricMetadataActionResult,
)

from ..container_metric import ContainerUtilizationMetricService


class ContainerUtilizationMetricProcessors(ActionProcessor):
    query_metadata: ActionProcessor[
        ContainerMetricMetadataAction, ContainerMetricMetadataActionResult
    ]
    query_metric: ActionProcessor[ContainerMetricAction, ContainerMetricActionResult]

    def __init__(
        self, service: ContainerUtilizationMetricService, action_monitors: list[ActionMonitor]
    ) -> None:
        self.query_metadata = ActionProcessor(service.query_metadata, action_monitors)
        self.query_metric = ActionProcessor(service.query_metric, action_monitors)



"""Agent selector interfaces and implementations for sokovan scheduler."""

# Import exceptions first (they don't have dependencies)
from .exceptions import (
    AgentSelectionError,
    NoAvailableAgentError,
    NoCompatibleAgentError,
)

# Then import selector which depends on exceptions
from .selector import (
    AbstractAgentSelector,
    AgentInfo,
    AgentSelection,
    AgentSelectionConfig,
    AgentSelectionCriteria,
    AgentSelector,
    KernelResourceSpec,
    ResourceRequirements,
    SessionMetadata,
)

__all__ = [
    # Exceptions
    "AgentSelectionError",
    "NoAvailableAgentError",
    "NoCompatibleAgentError",
    # Core interfaces
    "AbstractAgentSelector",
    "AgentInfo",
    "AgentSelection",
    "AgentSelectionConfig",
    "AgentSelectionCriteria",
    "AgentSelector",
    "KernelResourceSpec",
    "ResourceRequirements",
    "SessionMetadata",
]



"""Scheduler metrics for monitoring session scheduling performance."""

import time
from contextlib import contextmanager
from typing import Iterator, Optional, Self

from prometheus_client import Counter, Histogram


class SchedulerOperationMetricObserver:
    """Metrics for high-level scheduler operations (schedule, start, terminate, etc.)."""

    _instance: Optional[Self] = None

    _operation_count: Counter
    _operation_success_count: Counter
    _operation_duration_sec: Histogram

    def __init__(self) -> None:
        self._operation_count = Counter(
            name="backendai_scheduler_operation_count",
            documentation="Total number of scheduler operations",
            labelnames=["operation", "status"],
        )
        self._operation_success_count = Counter(
            name="backendai_scheduler_operation_success_count",
            documentation="Total number of successful scheduler operations",
            labelnames=["operation"],
        )
        self._operation_duration_sec = Histogram(
            name="backendai_scheduler_operation_duration_sec",
            documentation="Duration of scheduler operations in seconds",
            labelnames=["operation", "status"],
            buckets=[0.01, 0.05, 0.1, 0.5, 1.0, 2.5, 5.0, 10.0, 30.0],
        )

    @classmethod
    def instance(cls) -> Self:
        """Get or create the singleton instance."""
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def observe_operation(
        self,
        *,
        operation: str,
        status: str,
        duration: float,
    ) -> None:
        """
        Record a scheduler operation.

        Args:
            operation: The operation type (e.g., 'schedule_all_scaling_groups', 'start_session', 'terminate_session')
            status: The operation status ('success' or 'failure')
            duration: The operation duration in seconds
        """
        self._operation_count.labels(operation=operation, status=status).inc()
        self._operation_duration_sec.labels(operation=operation, status=status).observe(duration)

    @contextmanager
    def measure_operation(self, operation: str) -> Iterator[None]:
        """
        Context manager to measure a scheduler operation.

        Args:
            operation: The operation type
        """
        start = time.perf_counter()
        status = "success"
        try:
            yield
        except Exception:
            status = "failure"
            raise
        finally:
            duration = time.perf_counter() - start
            self.observe_operation(operation=operation, status=status, duration=duration)

    def observe_success(self, operation: str, count: int) -> None:
        """
        Record a successful operation.

        Args:
            operation: The operation type
            count: The number of successful operations
        """
        self._operation_success_count.labels(operation=operation).inc(count)


class SchedulerPhaseMetricObserver:
    """Metrics for scheduler and scheduling controller execution phases."""

    _instance: Optional[Self] = None

    _phase_count: Counter
    _phase_duration_sec: Histogram

    def __init__(self) -> None:
        self._phase_count = Counter(
            name="backendai_scheduler_phase_count",
            documentation="Total number of scheduler/scheduling controller phase executions",
            labelnames=["component", "scaling_group", "phase", "status"],
        )
        self._phase_duration_sec = Histogram(
            name="backendai_scheduler_phase_duration_sec",
            documentation="Duration of scheduler/scheduling controller phase executions in seconds",
            labelnames=["component", "scaling_group", "phase", "status"],
            buckets=[0.001, 0.005, 0.01, 0.05, 0.1, 0.5, 1.0, 2.5, 5.0, 10.0],
        )

    @classmethod
    def instance(cls) -> Self:
        """Get or create the singleton instance."""
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def _observe_phase(
        self,
        *,
        component: str,
        scaling_group: str,
        phase: str,
        status: str,
        duration: float,
    ) -> None:
        """
        Record a scheduler or scheduling controller phase execution.

        Args:
            component: The component name ('scheduler' or 'scheduling_controller')
            scaling_group: The scaling group name
            phase: The phase name (e.g., 'validation', 'sequencing_fifo', 'allocation', 'fetch_data')
            status: The phase status ('success' or 'failure')
            duration: The phase duration in seconds
        """
        self._phase_count.labels(
            component=component, scaling_group=scaling_group, phase=phase, status=status
        ).inc()
        self._phase_duration_sec.labels(
            component=component, scaling_group=scaling_group, phase=phase, status=status
        ).observe(duration)

    @contextmanager
    def measure_phase(self, component: str, scaling_group: str, phase: str) -> Iterator[None]:
        """
        Context manager to measure a phase execution.

        Args:
            component: The component name ('scheduler' or 'scheduling_controller')
            scaling_group: The scaling group name
            phase: The phase name
        """
        start = time.perf_counter()
        status = "success"
        try:
            yield
        except Exception:
            status = "failure"
            raise
        finally:
            duration = time.perf_counter() - start
            self._observe_phase(
                component=component,
                scaling_group=scaling_group,
                phase=phase,
                status=status,
                duration=duration,
            )



import logging

from ai.backend.common.exception import InvalidAPIParameters
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.models.user import UserRole
from ai.backend.manager.repositories.domain.admin_repository import AdminDomainRepository
from ai.backend.manager.repositories.domain.repository import DomainRepository
from ai.backend.manager.services.domain.actions.create_domain import (
    CreateDomainAction,
    CreateDomainActionResult,
)
from ai.backend.manager.services.domain.actions.create_domain_node import (
    CreateDomainNodeAction,
    CreateDomainNodeActionResult,
)
from ai.backend.manager.services.domain.actions.delete_domain import (
    DeleteDomainAction,
    DeleteDomainActionResult,
)
from ai.backend.manager.services.domain.actions.modify_domain import (
    ModifyDomainAction,
    ModifyDomainActionResult,
)
from ai.backend.manager.services.domain.actions.modify_domain_node import (
    ModifyDomainNodeAction,
    ModifyDomainNodeActionResult,
)
from ai.backend.manager.services.domain.actions.purge_domain import (
    PurgeDomainAction,
    PurgeDomainActionResult,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class DomainService:
    _repository: DomainRepository
    _admin_repository: AdminDomainRepository

    def __init__(
        self,
        repository: DomainRepository,
        admin_repository: AdminDomainRepository,
    ) -> None:
        self._repository = repository
        self._admin_repository = admin_repository

    async def create_domain(self, action: CreateDomainAction) -> CreateDomainActionResult:
        domain_name_candidate = action.creator.name.strip()
        if domain_name_candidate == "" or len(domain_name_candidate) > 64:
            raise InvalidAPIParameters("Domain name cannot be empty or exceed 64 characters.")

        if action.user_info.role == UserRole.SUPERADMIN:
            domain_data = await self._admin_repository.create_domain_force(action.creator)
        else:
            domain_data = await self._repository.create_domain_validated(action.creator)
        return CreateDomainActionResult(
            domain_data=domain_data,
        )

    async def modify_domain(self, action: ModifyDomainAction) -> ModifyDomainActionResult:
        if action.user_info.role == UserRole.SUPERADMIN:
            domain_data = await self._admin_repository.modify_domain_force(
                action.domain_name, action.modifier
            )
        else:
            domain_data = await self._repository.modify_domain_validated(
                action.domain_name, action.modifier
            )
        return ModifyDomainActionResult(
            domain_data=domain_data,
        )

    async def delete_domain(self, action: DeleteDomainAction) -> DeleteDomainActionResult:
        if action.user_info.role == UserRole.SUPERADMIN:
            await self._admin_repository.soft_delete_domain_force(action.name)
        else:
            await self._repository.soft_delete_domain_validated(action.name)

        return DeleteDomainActionResult(
            name=action.name,
        )

    async def purge_domain(self, action: PurgeDomainAction) -> PurgeDomainActionResult:
        name = action.name

        if action.user_info.role == UserRole.SUPERADMIN:
            await self._admin_repository.purge_domain_force(name)
        else:
            await self._repository.purge_domain_validated(name)
        return PurgeDomainActionResult(
            name=name,
        )

    async def create_domain_node(
        self, action: CreateDomainNodeAction
    ) -> CreateDomainNodeActionResult:
        domain_name_candidate = action.creator.name.strip()
        if domain_name_candidate == "" or len(domain_name_candidate) > 64:
            raise InvalidAPIParameters("Domain name cannot be empty or exceed 64 characters.")

        scaling_groups = action.scaling_groups

        if action.user_info.role == UserRole.SUPERADMIN:
            domain_data = await self._admin_repository.create_domain_node_with_permissions_force(
                action.creator, action.user_info, scaling_groups
            )
        else:
            domain_data = await self._repository.create_domain_node_with_permissions(
                action.creator, action.user_info, scaling_groups
            )

        return CreateDomainNodeActionResult(
            domain_data=domain_data,
        )

    async def modify_domain_node(
        self, action: ModifyDomainNodeAction
    ) -> ModifyDomainNodeActionResult:
        domain_name = action.name

        if action.sgroups_to_add is not None and action.sgroups_to_remove is not None:
            if union := action.sgroups_to_add | action.sgroups_to_remove:
                raise InvalidAPIParameters(
                    "Should be no scaling group names included in both `sgroups_to_add` and `sgroups_to_remove` "
                    f"(sg:{union})."
                )

        if action.user_info.role == UserRole.SUPERADMIN:
            domain_data = await self._admin_repository.modify_domain_node_with_permissions_force(
                domain_name,
                action.modifier.fields_to_update(),
                action.user_info,
                action.sgroups_to_add,
                action.sgroups_to_remove,
            )
        else:
            domain_data = await self._repository.modify_domain_node_with_permissions(
                domain_name,
                action.modifier.fields_to_update(),
                action.user_info,
                action.sgroups_to_add,
                action.sgroups_to_remove,
            )
        return ModifyDomainNodeActionResult(
            domain_data=domain_data,
        )



"""migrate-roundrobin-strategy-to-agent-selector-config

Revision ID: c4b7ec740b36
Revises: 59a622c31820
Create Date: 2024-09-17 00:31:31.379466

"""

from alembic import op
from sqlalchemy.sql import text

# revision identifiers, used by Alembic.
revision = "c4b7ec740b36"
down_revision = "59a622c31820"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute(
        text("""
    UPDATE scaling_groups
    SET scheduler_opts =
    CASE
        WHEN scheduler_opts ? 'roundrobin' AND scheduler_opts->>'roundrobin' = 'true' THEN
            jsonb_set(
                jsonb_set(
                    scheduler_opts - 'roundrobin',
                    '{agent_selection_strategy}', '"roundrobin"'::jsonb
                ),
                '{agent_selector_config}', '{}'::jsonb
            )
        ELSE jsonb_set(
            scheduler_opts - 'roundrobin',
            '{agent_selector_config}', '{}'::jsonb
        )
    END;
    """)
    )


def downgrade() -> None:
    op.execute(
        text("""
    UPDATE scaling_groups
    SET scheduler_opts =
    CASE
        WHEN scheduler_opts->>'agent_selection_strategy' = 'roundrobin' THEN
            jsonb_set(
                jsonb_set(
                    scheduler_opts - 'agent_selector_config',
                    '{agent_selection_strategy}', '"legacy"'::jsonb
                ),
                '{roundrobin}', 'true'::jsonb
            )
        ELSE scheduler_opts - 'agent_selector_config'
    END;
    """)
    )



from ai.backend.common.exception import ErrorCode, ErrorDetail, ErrorDomain, ErrorOperation
from ai.backend.manager.errors.common import ObjectNotFound


class DefinitionFileNotFound(ObjectNotFound):
    object_name = "definition-file"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.MODEL_SERVICE,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )



from dataclasses import dataclass
from typing import Any, Optional, override

from ai.backend.manager.types import Creator


@dataclass
class HuggingFaceRegistryCreator(Creator):
    url: str
    token: Optional[str]

    @override
    def fields_to_store(self) -> dict[str, Any]:
        return {
            "url": self.url,
            "token": self.token,
        }



import uuid
from dataclasses import dataclass, field
from enum import Enum
from typing import Optional

from ai.backend.common.data.artifact.types import ArtifactRegistryType
from ai.backend.manager.api.gql.base import IntFilter, StringFilter
from ai.backend.manager.data.artifact.types import (
    ArtifactAvailability,
    ArtifactOrderField,
    ArtifactRevisionOrderField,
    ArtifactStatus,
    ArtifactType,
)


@dataclass
class ArtifactOrderingOptions:
    """Ordering options for artifact queries."""

    order_by: list[tuple[ArtifactOrderField, bool]] = field(
        default_factory=lambda: [(ArtifactOrderField.NAME, False)]
    )  # (field, desc)


@dataclass
class ArtifactRevisionOrderingOptions:
    """Ordering options for artifact revision queries."""

    order_by: list[tuple[ArtifactRevisionOrderField, bool]] = field(
        default_factory=lambda: [(ArtifactRevisionOrderField.CREATED_AT, False)]
    )  # (field, desc)


@dataclass
class ArtifactFilterOptions:
    """Filtering options for artifacts."""

    artifact_type: Optional[list[ArtifactType]] = None
    name_filter: Optional[StringFilter] = None
    registry_filter: Optional[StringFilter] = None
    source_filter: Optional[StringFilter] = None
    registry_id: Optional[uuid.UUID] = None
    registry_type: Optional[ArtifactRegistryType] = None
    source_registry_id: Optional[uuid.UUID] = None
    source_registry_type: Optional[ArtifactRegistryType] = None
    availability: Optional[list[ArtifactAvailability]] = None

    # Logical operations
    AND: Optional[list["ArtifactFilterOptions"]] = None
    OR: Optional[list["ArtifactFilterOptions"]] = None
    NOT: Optional[list["ArtifactFilterOptions"]] = None


class ArtifactStatusFilterType(Enum):
    IN = "in"
    EQUALS = "equals"


@dataclass
class ArtifactStatusFilter:
    """Status filter with operation type and values."""

    type: ArtifactStatusFilterType
    values: list[ArtifactStatus]


@dataclass
class ArtifactRevisionFilterOptions:
    """Filtering options for artifact revisions."""

    artifact_id: Optional[uuid.UUID] = None
    status_filter: Optional[ArtifactStatusFilter] = None
    version_filter: Optional[StringFilter] = None
    size_filter: Optional[IntFilter] = None

    # Logical operations
    AND: Optional[list["ArtifactRevisionFilterOptions"]] = None
    OR: Optional[list["ArtifactRevisionFilterOptions"]] = None
    NOT: Optional[list["ArtifactRevisionFilterOptions"]] = None



from collections.abc import Collection
from typing import Callable, Optional

import sqlalchemy as sa

from ai.backend.common.types import AgentId
from ai.backend.manager.data.agent.types import AgentStatus
from ai.backend.manager.models.agent import AgentRow

type QueryConditionCallable = Callable[
    [Optional[sa.sql.expression.BinaryExpression]], sa.sql.expression.BinaryExpression
]
type QueryCondition = Callable[[sa.sql.Select], QueryConditionCallable]

type QueryOrder = sa.sql.ClauseElement


class QueryConditions:
    @staticmethod
    def by_ids(agent_ids: Collection[AgentId]) -> QueryCondition:
        def inner(stmt: sa.sql.Select) -> sa.sql.Select:
            cond = AgentRow.id.in_(agent_ids)
            return stmt.where(cond)

        return inner

    @staticmethod
    def by_scaling_group(scaling_group: str) -> QueryCondition:
        def inner(stmt: sa.sql.Select) -> sa.sql.Select:
            cond = AgentRow.scaling_group == scaling_group
            return stmt.where(cond)

        return inner

    @staticmethod
    def by_statuses(statuses: Collection[AgentStatus]) -> QueryCondition:
        def inner(stmt: sa.sql.Select) -> sa.sql.Select:
            cond = AgentRow.status.in_(statuses)
            return stmt.where(cond)

        return inner


class QueryOrders:
    @staticmethod
    def id(ascending: bool = True) -> QueryOrder:
        if ascending:
            return AgentRow.id.asc()
        else:
            return AgentRow.id.desc()

    @staticmethod
    def status(ascending: bool = True) -> QueryOrder:
        if ascending:
            return AgentRow.status.asc()
        else:
            return AgentRow.status.desc()

    @staticmethod
    def scaling_group(ascending: bool = True) -> QueryOrder:
        if ascending:
            return AgentRow.scaling_group.asc()
        else:
            return AgentRow.scaling_group.desc()



"""Types for database source operations."""

import sqlalchemy as sa
from sqlalchemy.ext.asyncio import AsyncSession as SASession

from ai.backend.common.types import SessionId
from ai.backend.manager.errors.kernel import SessionNotFound
from ai.backend.manager.models import SessionRow


class SessionRowCache:
    """Cache for pre-fetched session rows with lazy loading support."""

    def __init__(self, prefetched: dict[SessionId, SessionRow]):
        self._cache = prefetched

    async def get_or_fetch(self, db_sess: SASession, session_id: SessionId) -> SessionRow:
        """Get session from cache or fetch from database if not present."""
        if session_id in self._cache:
            return self._cache[session_id]

        # Fetch if not in cache
        query = sa.select(SessionRow).where(SessionRow.id == session_id)
        result = await db_sess.execute(query)
        session_row = result.scalar_one_or_none()

        if session_row is None:
            raise SessionNotFound(str(session_id))

        # Cache for future use
        self._cache[session_id] = session_row
        return session_row



import asyncio
import logging
import secrets
import uuid
from http import HTTPStatus

import aiohttp
import tomli
from pydantic import HttpUrl
from yarl import URL

from ai.backend.common.bgtask.bgtask import BackgroundTaskManager, ProgressReporter
from ai.backend.common.clients.valkey_client.valkey_live.client import ValkeyLiveClient
from ai.backend.common.events.dispatcher import (
    EventDispatcher,
    EventHandler,
)
from ai.backend.common.events.event_types.kernel.types import (
    KernelLifecycleEventReason,
)
from ai.backend.common.events.event_types.model_serving.broadcast import (
    ModelServiceStatusBroadcastEvent,
)
from ai.backend.common.events.event_types.session.broadcast import (
    SessionCancelledBroadcastEvent,
    SessionEnqueuedBroadcastEvent,
    SessionPreparingBroadcastEvent,
    SessionStartedBroadcastEvent,
    SessionTerminatedBroadcastEvent,
)
from ai.backend.common.json import dump_json_str
from ai.backend.common.types import (
    AgentId,
    ImageAlias,
    SessionTypes,
)
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.data.image.types import ImageIdentifier
from ai.backend.manager.data.model_serving.creator import EndpointCreator
from ai.backend.manager.data.model_serving.types import (
    CompactServiceInfo,
    ErrorInfo,
    ModelServiceDefinition,
    RequesterCtx,
    RouteInfo,
    ServiceInfo,
)
from ai.backend.manager.data.model_serving.types import (
    EndpointTokenData as ServiceEndpointTokenData,
)
from ai.backend.manager.errors.service import (
    EndpointNotFound,
    ModelServiceNotFound,
    RouteNotFound,
)
from ai.backend.manager.errors.storage import UnexpectedStorageProxyResponseError
from ai.backend.manager.models.endpoint import (
    EndpointLifecycle,
    EndpointTokenRow,
)
from ai.backend.manager.models.routing import RouteStatus
from ai.backend.manager.models.session import KernelLoadingStrategy
from ai.backend.manager.models.storage import StorageSessionManager
from ai.backend.manager.models.user import UserRole
from ai.backend.manager.models.vfolder import VFolderOwnershipType, VFolderRow
from ai.backend.manager.registry import AgentRegistry
from ai.backend.manager.repositories.model_serving.admin_repository import (
    AdminModelServingRepository,
)
from ai.backend.manager.repositories.model_serving.repository import ModelServingRepository
from ai.backend.manager.services.model_serving.actions.clear_error import (
    ClearErrorAction,
    ClearErrorActionResult,
)
from ai.backend.manager.services.model_serving.actions.create_model_service import (
    CreateModelServiceAction,
    CreateModelServiceActionResult,
)
from ai.backend.manager.services.model_serving.actions.delete_model_service import (
    DeleteModelServiceAction,
    DeleteModelServiceActionResult,
)
from ai.backend.manager.services.model_serving.actions.delete_route import (
    DeleteRouteAction,
    DeleteRouteActionResult,
)
from ai.backend.manager.services.model_serving.actions.dry_run_model_service import (
    DryRunModelServiceAction,
    DryRunModelServiceActionResult,
)
from ai.backend.manager.services.model_serving.actions.force_sync import (
    ForceSyncAction,
    ForceSyncActionResult,
)
from ai.backend.manager.services.model_serving.actions.generate_token import (
    GenerateTokenAction,
    GenerateTokenActionResult,
)
from ai.backend.manager.services.model_serving.actions.get_model_service_info import (
    GetModelServiceInfoAction,
    GetModelServiceInfoActionResult,
)
from ai.backend.manager.services.model_serving.actions.list_errors import (
    ListErrorsAction,
    ListErrorsActionResult,
)
from ai.backend.manager.services.model_serving.actions.list_model_service import (
    ListModelServiceAction,
    ListModelServiceActionResult,
)
from ai.backend.manager.services.model_serving.actions.modify_endpoint import (
    ModifyEndpointAction,
    ModifyEndpointActionResult,
)
from ai.backend.manager.services.model_serving.actions.update_route import (
    UpdateRouteAction,
    UpdateRouteActionResult,
)
from ai.backend.manager.services.model_serving.exceptions import (
    GenericForbidden,
    InvalidAPIParameters,
)
from ai.backend.manager.sokovan.deployment.deployment_controller import DeploymentController
from ai.backend.manager.sokovan.deployment.types import DeploymentLifecycleType
from ai.backend.manager.types import UserScope

log = BraceStyleAdapter(logging.getLogger(__name__))


class ModelServingService:
    _agent_registry: AgentRegistry
    _background_task_manager: BackgroundTaskManager
    _event_dispatcher: EventDispatcher
    _storage_manager: StorageSessionManager
    _config_provider: ManagerConfigProvider
    _repository: ModelServingRepository
    _admin_repository: AdminModelServingRepository

    _valkey_live: ValkeyLiveClient
    _deployment_controller: DeploymentController

    def __init__(
        self,
        agent_registry: AgentRegistry,
        background_task_manager: BackgroundTaskManager,
        event_dispatcher: EventDispatcher,
        storage_manager: StorageSessionManager,
        config_provider: ManagerConfigProvider,
        valkey_live: ValkeyLiveClient,
        repository: ModelServingRepository,
        admin_repository: AdminModelServingRepository,
        deployment_controller: DeploymentController,
    ) -> None:
        self._agent_registry = agent_registry
        self._background_task_manager = background_task_manager
        self._event_dispatcher = event_dispatcher
        self._storage_manager = storage_manager
        self._config_provider = config_provider
        self._valkey_live = valkey_live
        self._repository = repository
        self._admin_repository = admin_repository
        self._deployment_controller = deployment_controller

    async def _fetch_file_from_storage_proxy(
        self,
        filename: str,
        model_vfolder_row: VFolderRow,
    ) -> bytes:
        vfid = model_vfolder_row.vfid
        folder_host = model_vfolder_row.host

        proxy_name, volume_name = self._storage_manager.get_proxy_and_volume(folder_host)

        manager_client = self._storage_manager.get_manager_facing_client(proxy_name)
        return await manager_client.fetch_file_content(
            volume_name,
            str(vfid),
            f"./{filename}",
        )

    async def create(self, action: CreateModelServiceAction) -> CreateModelServiceActionResult:
        service_prepare_ctx = action.creator.model_service_prepare_ctx

        # Get model vfolder
        model_vfolder_row = await self._repository.get_vfolder_by_id(service_prepare_ctx.model_id)
        if not model_vfolder_row:
            raise InvalidAPIParameters("Model vfolder not found")
        if model_vfolder_row.ownership_type == VFolderOwnershipType.GROUP:
            raise InvalidAPIParameters(
                "Cannot create model service with the project type's vfolder"
            )

        try:
            chunks = await self._fetch_file_from_storage_proxy(
                "service-definition.toml", model_vfolder_row
            )
        except UnexpectedStorageProxyResponseError:
            chunks = None

        if chunks:
            raw_service_definition = chunks.decode("utf-8")
            service_definition = tomli.loads(raw_service_definition)

            definition = action.creator.runtime_variant
            if definition in service_definition:
                variant_def = ModelServiceDefinition.model_validate(service_definition[definition])
                if variant_def.resource_slots:
                    action.creator.config.resources = variant_def.resource_slots
                if variant_def.environment:
                    action.creator.image = variant_def.environment.image
                    action.creator.architecture = variant_def.environment.architecture
                if variant_def.environ:
                    if action.creator.config.environ:
                        action.creator.config.environ.update(variant_def.environ)
                    else:
                        action.creator.config.environ = variant_def.environ

        # Resolve image row - EndpointRow constructor needs the actual ImageRow object
        image_row = await self._repository.resolve_image_for_endpoint_creation([
            ImageIdentifier(action.creator.image, action.creator.architecture),
            ImageAlias(action.creator.image),
        ])

        creation_config = action.creator.config.to_dict()
        creation_config["mounts"] = [
            service_prepare_ctx.model_id,
            *[m.vfid.folder_id for m in service_prepare_ctx.extra_mounts],
        ]
        creation_config["mount_map"] = {
            service_prepare_ctx.model_id: action.creator.config.model_mount_destination,
            **{
                m.vfid.folder_id: m.kernel_path.as_posix() for m in service_prepare_ctx.extra_mounts
            },
        }
        creation_config["mount_options"] = {
            m.vfid.folder_id: {"permission": m.mount_perm} for m in service_prepare_ctx.extra_mounts
        }
        sudo_session_enabled = action.creator.sudo_session_enabled

        # check if session is valid to be created
        await self._agent_registry.create_session(
            "",
            image_row.image_ref,
            UserScope(
                domain_name=action.creator.domain_name,
                group_id=service_prepare_ctx.group_id,
                user_uuid=service_prepare_ctx.owner_uuid,
                user_role=service_prepare_ctx.owner_role,
            ),
            service_prepare_ctx.owner_access_key,
            service_prepare_ctx.resource_policy,
            SessionTypes.INFERENCE,
            creation_config,
            action.creator.cluster_mode,
            action.creator.cluster_size,
            dry_run=True,  # Setting this to True will prevent actual session from being enqueued
            bootstrap_script=action.creator.bootstrap_script,
            startup_command=action.creator.startup_command,
            tag=action.creator.tag,
            callback_url=URL(action.creator.callback_url.unicode_string())
            if action.creator.callback_url
            else None,
            sudo_session_enabled=sudo_session_enabled,
        )

        # Check endpoint name uniqueness
        is_name_available = await self._repository.check_endpoint_name_uniqueness(
            action.creator.service_name
        )
        if not is_name_available:
            raise InvalidAPIParameters("Cannot create multiple services with same name")

        project_id = await self._repository.resolve_group_id(
            action.creator.domain_name, action.creator.group_name
        )
        if project_id is None:
            raise InvalidAPIParameters(f"Invalid group name {action.creator.group_name}")

        endpoint = EndpointCreator(
            action.creator.service_name,
            model_definition_path=service_prepare_ctx.model_definition_path,
            created_user=action.request_user_id,
            session_owner=service_prepare_ctx.owner_uuid,
            replicas=action.creator.replicas,
            image=image_row.id,
            model=service_prepare_ctx.model_id,
            domain=action.creator.domain_name,
            project=project_id,
            resource_group=service_prepare_ctx.scaling_group,
            resource_slots=action.creator.config.resources,
            cluster_mode=action.creator.cluster_mode,
            cluster_size=action.creator.cluster_size,
            extra_mounts=list(service_prepare_ctx.extra_mounts),
            model_mount_destination=action.creator.config.model_mount_destination,
            tag=action.creator.tag,
            startup_command=action.creator.startup_command,
            callback_url=URL(action.creator.callback_url.unicode_string())
            if action.creator.callback_url
            else None,
            environ=action.creator.config.environ,
            bootstrap_script=action.creator.bootstrap_script,
            resource_opts=action.creator.config.resource_opts,
            open_to_public=action.creator.open_to_public,
            runtime_variant=action.creator.runtime_variant,
        )

        endpoint_data = await self._repository.create_endpoint_validated(
            endpoint, self._agent_registry
        )
        endpoint_id = endpoint_data.id

        return CreateModelServiceActionResult(
            ServiceInfo(
                endpoint_id=endpoint_id,
                model_id=endpoint.model,
                extra_mounts=[m.vfid.folder_id for m in endpoint.extra_mounts],
                name=action.creator.service_name,
                model_definition_path=service_prepare_ctx.model_definition_path,
                replicas=endpoint.replicas,
                desired_session_count=endpoint.replicas,
                active_routes=[],
                service_endpoint=None,
                is_public=action.creator.open_to_public,
                runtime_variant=action.creator.runtime_variant,
            )
        )

    async def list_serve(self, action: ListModelServiceAction) -> ListModelServiceActionResult:
        endpoints = await self._repository.list_endpoints_by_owner_validated(
            action.session_owener_id, name=action.name
        )

        return ListModelServiceActionResult(
            data=[
                CompactServiceInfo(
                    id=endpoint.id,
                    name=endpoint.name,
                    replicas=endpoint.replicas,
                    desired_session_count=endpoint.replicas,
                    active_route_count=len([
                        r for r in endpoint.routings if r.status == RouteStatus.HEALTHY
                    ])
                    if endpoint.routings
                    else 0,
                    service_endpoint=HttpUrl(endpoint.url) if endpoint.url else None,
                    is_public=endpoint.open_to_public,
                )
                for endpoint in endpoints
            ]
        )

    async def check_requester_access(self, requester_ctx: RequesterCtx) -> None:
        if requester_ctx.is_authorized is False:
            raise GenericForbidden("Only authorized requests may have access key scopes.")

    async def delete(self, action: DeleteModelServiceAction) -> DeleteModelServiceActionResult:
        service_id = action.service_id

        # Get endpoint with access validation
        await self.check_requester_access(action.requester_ctx)
        if action.requester_ctx.user_role == UserRole.SUPERADMIN:
            endpoint_data = await self._admin_repository.get_endpoint_by_id_force(service_id)
        else:
            endpoint_data = await self._repository.get_endpoint_by_id_validated(
                service_id,
                action.requester_ctx.user_id,
                action.requester_ctx.user_role,
                action.requester_ctx.domain_name,
            )

        if not endpoint_data:
            raise ModelServiceNotFound

        # Determine lifecycle stage based on routes
        has_routes = endpoint_data.routings and len(endpoint_data.routings) > 0
        if has_routes:
            lifecycle_stage = EndpointLifecycle.DESTROYING
            replicas = 0
        else:
            lifecycle_stage = EndpointLifecycle.DESTROYED
            replicas = None

        # Update endpoint lifecycle
        if action.requester_ctx.user_role == UserRole.SUPERADMIN:
            await self._admin_repository.update_endpoint_lifecycle_force(
                service_id, lifecycle_stage, replicas
            )
        else:
            await self._repository.update_endpoint_lifecycle_validated(
                service_id,
                lifecycle_stage,
                action.requester_ctx.user_id,
                action.requester_ctx.user_role,
                action.requester_ctx.domain_name,
                replicas,
            )

        return DeleteModelServiceActionResult(success=True)

    async def dry_run(self, action: DryRunModelServiceAction) -> DryRunModelServiceActionResult:
        # TODO: Seperate background task definition and trigger into different layer
        service_prepare_ctx = action.model_service_prepare_ctx
        # Get user with keypair
        created_user = await self._repository.get_user_with_keypair(action.request_user_id)
        if not created_user:
            raise InvalidAPIParameters("User not found")

        image_row = await self._repository.resolve_image_for_endpoint_creation([
            ImageIdentifier(action.image, action.architecture),
            ImageAlias(action.image),
        ])

        creation_config = action.config.to_dict()
        creation_config["mount_map"] = {
            service_prepare_ctx.model_id: action.config.model_mount_destination
        }
        sudo_session_enabled = action.sudo_session_enabled

        async def _task(reporter: ProgressReporter) -> None:
            terminated_event = asyncio.Event()

            result = await self._agent_registry.create_session(
                f"model-eval-{secrets.token_urlsafe(16)}",
                image_row.image_ref,
                UserScope(
                    domain_name=action.domain_name,
                    group_id=service_prepare_ctx.group_id,
                    user_uuid=created_user.uuid,
                    user_role=created_user.role,
                ),
                service_prepare_ctx.owner_access_key,
                service_prepare_ctx.resource_policy,
                SessionTypes.INFERENCE,
                {
                    "mounts": [
                        service_prepare_ctx.model_id,
                        *[m.vfid for m in service_prepare_ctx.extra_mounts],
                    ],
                    "mount_map": {
                        service_prepare_ctx.model_id: action.config.model_mount_destination,
                        **{m.vfid: m.kernel_path for m in service_prepare_ctx.extra_mounts},
                    },
                    "mount_options": {
                        m.vfid: {"permission": m.mount_perm}
                        for m in service_prepare_ctx.extra_mounts
                    },
                    "model_definition_path": service_prepare_ctx.model_definition_path,
                    "environ": creation_config["environ"],
                    "scaling_group": service_prepare_ctx.scaling_group,
                    "resources": creation_config["resources"],
                    "resource_opts": creation_config["resource_opts"],
                    "preopen_ports": None,
                    "agent_list": None,
                },
                action.cluster_mode,
                action.cluster_size,
                bootstrap_script=action.bootstrap_script,
                startup_command=action.startup_command,
                tag=action.tag,
                callback_url=URL(action.callback_url.unicode_string())
                if action.callback_url
                else None,
                enqueue_only=True,
                sudo_session_enabled=sudo_session_enabled,
            )

            await reporter.update(
                message=dump_json_str({
                    "event": "session_enqueued",
                    "session_id": str(result["sessionId"]),
                })
            )

            async def _handle_event(
                _context: None,
                _source: AgentId,
                event: SessionEnqueuedBroadcastEvent
                | SessionPreparingBroadcastEvent
                | SessionStartedBroadcastEvent
                | SessionCancelledBroadcastEvent
                | SessionTerminatedBroadcastEvent
                | ModelServiceStatusBroadcastEvent,
            ) -> None:
                task_message = {"event": event.event_name(), "session_id": str(event.session_id)}
                match event:
                    case ModelServiceStatusBroadcastEvent():
                        task_message["is_healthy"] = event.new_status.value
                await reporter.update(message=dump_json_str(task_message))

                match event:
                    case SessionTerminatedBroadcastEvent() | SessionCancelledBroadcastEvent():
                        terminated_event.set()
                    case ModelServiceStatusBroadcastEvent():
                        session = await self._repository.get_session_by_id(
                            result["sessionId"],
                            KernelLoadingStrategy.ALL_KERNELS,
                        )
                        if not session:
                            log.warning(
                                "Session {} not found for event {}",
                                result["sessionId"],
                                event.event_name(),
                            )
                            return
                        await self._agent_registry.destroy_session(
                            session,
                            forced=True,
                        )

            session_event_matcher = lambda args: args[0] == str(result["sessionId"])
            model_service_event_matcher = lambda args: args[1] == str(result["sessionId"])

            handlers: list[EventHandler] = [
                self._event_dispatcher.subscribe(
                    SessionPreparingBroadcastEvent,
                    None,
                    _handle_event,
                    args_matcher=session_event_matcher,
                ),
                self._event_dispatcher.subscribe(
                    SessionStartedBroadcastEvent,
                    None,
                    _handle_event,
                    args_matcher=session_event_matcher,
                ),
                self._event_dispatcher.subscribe(
                    SessionCancelledBroadcastEvent,
                    None,
                    _handle_event,
                    args_matcher=session_event_matcher,
                ),
                self._event_dispatcher.subscribe(
                    SessionTerminatedBroadcastEvent,
                    None,
                    _handle_event,
                    args_matcher=session_event_matcher,
                ),
                self._event_dispatcher.subscribe(
                    ModelServiceStatusBroadcastEvent,
                    None,
                    _handle_event,
                    args_matcher=model_service_event_matcher,
                ),
            ]

            try:
                await terminated_event.wait()
            finally:
                for handler in handlers:
                    self._event_dispatcher.unsubscribe(handler)

        task_id = await self._background_task_manager.start(_task)
        return DryRunModelServiceActionResult(task_id)

    async def get_model_service_info(
        self, action: GetModelServiceInfoAction
    ) -> GetModelServiceInfoActionResult:
        # Get endpoint with access validation
        await self.check_requester_access(action.requester_ctx)
        if action.requester_ctx.user_role == UserRole.SUPERADMIN:
            endpoint_data = await self._admin_repository.get_endpoint_by_id_force(action.service_id)
        else:
            endpoint_data = await self._repository.get_endpoint_by_id_validated(
                action.service_id,
                action.requester_ctx.user_id,
                action.requester_ctx.user_role,
                action.requester_ctx.domain_name,
            )

        if not endpoint_data:
            raise ModelServiceNotFound

        return GetModelServiceInfoActionResult(
            ServiceInfo(
                endpoint_id=endpoint_data.id,
                model_id=endpoint_data.model,
                extra_mounts=[m.vfid.folder_id for m in endpoint_data.extra_mounts],
                name=endpoint_data.name,
                model_definition_path=endpoint_data.model_definition_path,
                replicas=endpoint_data.replicas,
                desired_session_count=endpoint_data.replicas,
                active_routes=[
                    RouteInfo(
                        route_id=r.id,
                        session_id=r.session,
                        traffic_ratio=r.traffic_ratio,
                    )
                    for r in endpoint_data.routings
                ]
                if endpoint_data.routings
                else [],
                service_endpoint=HttpUrl(endpoint_data.url) if endpoint_data.url else None,
                is_public=endpoint_data.open_to_public,
                runtime_variant=endpoint_data.runtime_variant,
            )
        )

    async def list_errors(self, action: ListErrorsAction) -> ListErrorsActionResult:
        # Get endpoint with access validation
        await self.check_requester_access(action.requester_ctx)
        if action.requester_ctx.user_role == UserRole.SUPERADMIN:
            endpoint_data = await self._admin_repository.get_endpoint_by_id_force(action.service_id)
        else:
            endpoint_data = await self._repository.get_endpoint_by_id_validated(
                action.service_id,
                action.requester_ctx.user_id,
                action.requester_ctx.user_role,
                action.requester_ctx.domain_name,
            )

        if not endpoint_data:
            raise ModelServiceNotFound

        error_routes = (
            [r for r in endpoint_data.routings if r.status == RouteStatus.FAILED_TO_START]
            if endpoint_data.routings
            else []
        )

        return ListErrorsActionResult(
            error_info=[
                ErrorInfo(
                    session_id=route.error_data.get("session_id"), error=route.error_data["errors"]
                )
                for route in error_routes
            ],
            retries=endpoint_data.retries,
        )

    async def clear_error(self, action: ClearErrorAction) -> ClearErrorActionResult:
        # Clear errors with access validation
        await self.check_requester_access(action.requester_ctx)
        if action.requester_ctx.user_role == UserRole.SUPERADMIN:
            success = await self._admin_repository.clear_endpoint_errors_force(action.service_id)
        else:
            success = await self._repository.clear_endpoint_errors_validated(
                action.service_id,
                action.requester_ctx.user_id,
                action.requester_ctx.user_role,
                action.requester_ctx.domain_name,
            )

        if not success:
            raise ModelServiceNotFound

        return ClearErrorActionResult(success=True)

    async def update_route(self, action: UpdateRouteAction) -> UpdateRouteActionResult:
        # Update route traffic with access validation
        await self.check_requester_access(action.requester_ctx)
        if action.requester_ctx.user_role == UserRole.SUPERADMIN:
            endpoint_data = await self._admin_repository.update_route_traffic_force(
                self._valkey_live, action.route_id, action.service_id, action.traffic_ratio
            )
        else:
            endpoint_data = await self._repository.update_route_traffic_validated(
                self._valkey_live,
                action.route_id,
                action.service_id,
                action.traffic_ratio,
                action.requester_ctx.user_id,
                action.requester_ctx.user_role,
                action.requester_ctx.domain_name,
            )
        if not endpoint_data:
            raise ModelServiceNotFound

        await self._agent_registry.notify_endpoint_route_update_to_appproxy(endpoint_data.id)

        return UpdateRouteActionResult(success=True)

    async def delete_route(self, action: DeleteRouteAction) -> DeleteRouteActionResult:
        # Get route with access validation
        await self.check_requester_access(action.requester_ctx)
        if action.requester_ctx.user_role == UserRole.SUPERADMIN:
            route_data = await self._admin_repository.get_route_by_id_force(
                action.route_id, action.service_id
            )
        else:
            route_data = await self._repository.get_route_by_id_validated(
                action.route_id,
                action.service_id,
                action.requester_ctx.user_id,
                action.requester_ctx.user_role,
                action.requester_ctx.domain_name,
            )

        if not route_data:
            raise RouteNotFound

        if route_data.status == RouteStatus.PROVISIONING:
            raise InvalidAPIParameters("Cannot remove route in PROVISIONING status")

        # Get session for destruction
        route_row = await self._repository.get_route_with_session(action.route_id)
        if not route_row:
            raise RouteNotFound

        await self._agent_registry.destroy_session(
            route_row.session_row,
            forced=False,
            reason=KernelLifecycleEventReason.SERVICE_SCALED_DOWN,
        )

        # Decrease endpoint replicas
        await self.check_requester_access(action.requester_ctx)
        if action.requester_ctx.user_role == UserRole.SUPERADMIN:
            await self._admin_repository.decrease_endpoint_replicas_force(action.service_id)
        else:
            await self._repository.decrease_endpoint_replicas_validated(
                action.service_id,
                action.requester_ctx.user_id,
                action.requester_ctx.user_role,
                action.requester_ctx.domain_name,
            )

        return DeleteRouteActionResult(success=True)

    async def generate_token(self, action: GenerateTokenAction) -> GenerateTokenActionResult:
        # Get endpoint with access validation
        await self.check_requester_access(action.requester_ctx)
        if action.requester_ctx.user_role == UserRole.SUPERADMIN:
            endpoint_data = await self._admin_repository.get_endpoint_by_id_force(action.service_id)
        else:
            endpoint_data = await self._repository.get_endpoint_by_id_validated(
                action.service_id,
                action.requester_ctx.user_id,
                action.requester_ctx.user_role,
                action.requester_ctx.domain_name,
            )

        if not endpoint_data:
            raise ModelServiceNotFound

        # Get scaling group info
        scaling_group_data = await self._repository.get_scaling_group_info(
            endpoint_data.resource_group
        )
        if not scaling_group_data:
            raise InvalidAPIParameters(f"Scaling group {endpoint_data.resource_group} not found")

        # Generate token via wsproxy
        body = {"user_uuid": str(endpoint_data.session_owner_id), "exp": action.expires_at}
        async with aiohttp.ClientSession() as session:
            async with session.post(
                f"{scaling_group_data.wsproxy_addr}/v2/endpoints/{endpoint_data.id}/token",
                json=body,
                headers={
                    "accept": "application/json",
                    "X-BackendAI-Token": scaling_group_data.wsproxy_api_token,
                },
            ) as resp:
                resp_json = await resp.json()
                if resp.status != HTTPStatus.OK:
                    raise EndpointNotFound(
                        f"Failed to generate token: {resp.status} {resp.reason} {resp_json}"
                    )
                token = resp_json["token"]

        # Create token in database
        token_id = uuid.uuid4()
        token_row = EndpointTokenRow(
            token_id,
            token,
            endpoint_data.id,
            endpoint_data.domain,
            endpoint_data.project,
            endpoint_data.session_owner_id,
        )

        await self.check_requester_access(action.requester_ctx)
        if action.requester_ctx.user_role == UserRole.SUPERADMIN:
            token_data = await self._admin_repository.create_endpoint_token_force(token_row)
        else:
            token_data = await self._repository.create_endpoint_token_validated(
                token_row,
                action.requester_ctx.user_id,
                action.requester_ctx.user_role,
                action.requester_ctx.domain_name,
            )

        if not token_data:
            raise ModelServiceNotFound

        # Convert data types
        service_token_data = ServiceEndpointTokenData(
            id=token_data.id,
            token=token_data.token,
            endpoint=token_data.endpoint,
            session_owner=token_data.session_owner,
            domain=token_data.domain,
            project=token_data.project,
            created_at=token_data.created_at,
        )
        return GenerateTokenActionResult(data=service_token_data)

    async def force_sync_with_app_proxy(self, action: ForceSyncAction) -> ForceSyncActionResult:
        # Get endpoint with access validation
        await self.check_requester_access(action.requester_ctx)
        if action.requester_ctx.user_role == UserRole.SUPERADMIN:
            endpoint_data = await self._admin_repository.get_endpoint_by_id_force(action.service_id)
        else:
            endpoint_data = await self._repository.get_endpoint_by_id_validated(
                action.service_id,
                action.requester_ctx.user_id,
                action.requester_ctx.user_role,
                action.requester_ctx.domain_name,
            )
        if not endpoint_data:
            raise ModelServiceNotFound

        await self._agent_registry.notify_endpoint_route_update_to_appproxy(endpoint_data.id)

        return ForceSyncActionResult(success=True)

    async def modify_endpoint(self, action: ModifyEndpointAction) -> ModifyEndpointActionResult:
        result = await self._repository.modify_endpoint(
            action,
            self._agent_registry,
            self._config_provider.legacy_etcd_config_loader,
            self._storage_manager,
        )
        if action.modifier.replica_count_modified():
            # Notify appproxy to update routing info
            await self._deployment_controller.mark_lifecycle_needed(
                DeploymentLifecycleType.CHECK_REPLICA,
            )
        return ModifyEndpointActionResult(success=result.success, data=result.data)



from __future__ import annotations

import logging

import sqlalchemy as sa
from sqlalchemy.orm import foreign, relationship

from ai.backend.common.exception import RelationNotLoadedError
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.huggingface_registry.types import HuggingFaceRegistryData

from .base import (
    Base,
    IDColumn,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

__all__ = ("HuggingFaceRegistryRow",)


def _get_registry_artifact_join_condition():
    from ai.backend.manager.models.artifact import ArtifactRow

    return HuggingFaceRegistryRow.id == foreign(ArtifactRow.registry_id)


def _get_registry_meta_join_condition():
    from ai.backend.manager.models.artifact_registries import ArtifactRegistryRow

    return HuggingFaceRegistryRow.id == foreign(ArtifactRegistryRow.registry_id)


class HuggingFaceRegistryRow(Base):
    __tablename__ = "huggingface_registries"

    id = IDColumn("id")
    url = sa.Column("url", sa.String, nullable=False)
    token = sa.Column("token", sa.String, nullable=True, default=None)

    artifacts = relationship(
        "ArtifactRow",
        back_populates="huggingface_registry",
        primaryjoin=_get_registry_artifact_join_condition,
        viewonly=True,
    )
    meta = relationship(
        "ArtifactRegistryRow",
        back_populates="huggingface_registries",
        primaryjoin=_get_registry_meta_join_condition,
        uselist=False,
        viewonly=True,
    )

    def __str__(self) -> str:
        return f"HuggingFaceRegistryRow(id={self.id}, url={self.url}, token={self.token})"

    def __repr__(self) -> str:
        return self.__str__()

    def to_dataclass(self) -> HuggingFaceRegistryData:
        try:
            return HuggingFaceRegistryData(
                id=self.id, name=self.meta.name, url=self.url, token=self.token
            )
        except Exception:
            raise RelationNotLoadedError()



from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseAction


@dataclass
class VFolderAction(BaseAction):
    @override
    @classmethod
    def entity_type(cls) -> str:
        return "vfolder"



import logging
from decimal import Decimal
from typing import Mapping, Optional
from uuid import UUID

import trafaret as t

from ai.backend.common.clients.valkey_client.valkey_stat.client import ValkeyStatClient
from ai.backend.common.metrics.metric import LayerType
from ai.backend.common.types import AccessKey, ResourceSlot
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.data.resource_preset.types import ResourcePresetData
from ai.backend.manager.decorators.repository_decorator import (
    create_layer_aware_repository_decorator,
)
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine
from ai.backend.manager.services.resource_preset.types import (
    ResourcePresetCreator,
    ResourcePresetModifier,
)

from .cache_source.cache_source import ResourcePresetCacheSource
from .db_source.db_source import ResourcePresetDBSource
from .types import CheckPresetsResult
from .utils import suppress_with_log

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

# Layer-specific decorator for resource_preset repository
repository_decorator = create_layer_aware_repository_decorator(LayerType.RESOURCE_PRESET)


class ResourcePresetRepository:
    """Repository that orchestrates between DB and cache sources for resource preset operations."""

    _db_source: ResourcePresetDBSource
    _cache_source: ResourcePresetCacheSource
    _config_provider: ManagerConfigProvider

    def __init__(
        self,
        db: ExtendedAsyncSAEngine,
        valkey_stat: ValkeyStatClient,
        config_provider: ManagerConfigProvider,
    ) -> None:
        self._db_source = ResourcePresetDBSource(db)
        self._cache_source = ResourcePresetCacheSource(valkey_stat)
        self._config_provider = config_provider

    @repository_decorator()
    async def create_preset_validated(self, creator: ResourcePresetCreator) -> ResourcePresetData:
        """
        Creates a new resource preset.
        Raises ResourcePresetConflict if a preset with the same name and scaling group already exists.
        """
        preset = await self._db_source.create_preset(creator)
        with suppress_with_log(
            [Exception], message="Failed to invalidate cache after preset creation"
        ):
            await self._cache_source.invalidate_all_presets()
        return preset

    @repository_decorator()
    async def get_preset_by_id(self, preset_id: UUID) -> ResourcePresetData:
        """
        Gets a resource preset by ID.
        Raises ResourcePresetNotFound if the preset doesn't exist.
        """
        # Try cache first
        with suppress_with_log([Exception], message=f"Failed to get preset {preset_id} from cache"):
            preset = await self._cache_source.get_preset_by_id(preset_id)
            if preset:
                return preset

        # Fallback to DB
        preset = await self._db_source.get_preset_by_id(preset_id)
        with suppress_with_log([Exception], message=f"Failed to cache preset {preset_id}"):
            await self._cache_source.set_preset(preset)
        return preset

    @repository_decorator()
    async def get_preset_by_name(self, name: str) -> ResourcePresetData:
        """
        Gets a resource preset by name.
        Raises ResourcePresetNotFound if the preset doesn't exist.
        """
        # Try cache first
        with suppress_with_log([Exception], message=f"Failed to get preset '{name}' from cache"):
            preset = await self._cache_source.get_preset_by_name(name)
            if preset:
                return preset

        # Fallback to DB
        preset = await self._db_source.get_preset_by_name(name)
        with suppress_with_log([Exception], message=f"Failed to cache preset '{name}'"):
            await self._cache_source.set_preset(preset)
        return preset

    @repository_decorator()
    async def get_preset_by_id_or_name(
        self, preset_id: Optional[UUID], name: Optional[str]
    ) -> ResourcePresetData:
        """
        Gets a resource preset by ID or name.
        ID takes precedence if both are provided.
        Raises ResourcePresetNotFound if the preset doesn't exist.
        """
        return await self._db_source.get_preset_by_id_or_name(preset_id, name)

    @repository_decorator()
    async def modify_preset_validated(
        self, preset_id: Optional[UUID], name: Optional[str], modifier: ResourcePresetModifier
    ) -> ResourcePresetData:
        """
        Modifies an existing resource preset.
        Raises ResourcePresetNotFound if the preset doesn't exist.
        """
        preset = await self._db_source.modify_preset(preset_id, name, modifier)
        with suppress_with_log(
            [Exception], message="Failed to invalidate cache after preset modification"
        ):
            await self._cache_source.invalidate_preset(preset_id, name)
        return preset

    @repository_decorator()
    async def delete_preset_validated(
        self, preset_id: Optional[UUID], name: Optional[str]
    ) -> ResourcePresetData:
        """
        Deletes a resource preset.
        Returns the deleted preset data.
        Raises ObjectNotFound if the preset doesn't exist.
        """
        preset = await self._db_source.delete_preset(preset_id, name)
        with suppress_with_log(
            [Exception], message="Failed to invalidate cache after preset deletion"
        ):
            await self._cache_source.invalidate_preset(preset_id, name)
        return preset

    @repository_decorator()
    async def list_presets(
        self, scaling_group_name: Optional[str] = None
    ) -> list[ResourcePresetData]:
        """
        Lists all resource presets.
        If scaling_group_name is provided, returns presets for that scaling group and global presets.
        """
        # Try cache first
        with suppress_with_log([Exception], message="Failed to get preset list from cache"):
            presets = await self._cache_source.get_preset_list(scaling_group_name)
            if presets is not None:
                return presets

        # Fallback to DB
        await self._config_provider.legacy_etcd_config_loader.get_resource_slots()
        presets = await self._db_source.list_presets(scaling_group_name)

        # Cache the result
        with suppress_with_log([Exception], message="Failed to cache preset list"):
            await self._cache_source.set_preset_list(presets, scaling_group_name)

        return presets

    @repository_decorator()
    async def check_presets(
        self,
        access_key: AccessKey,
        user_id: UUID,
        group_name: str,
        domain_name: str,
        resource_policy: Mapping[str, str],
        scaling_group: Optional[str] = None,
    ) -> CheckPresetsResult:
        """
        Check resource presets availability and resource limits.
        """
        # Get configuration values
        known_slot_types = (
            await self._config_provider.legacy_etcd_config_loader.get_resource_slots()
        )
        # Try to get from cache first
        with suppress_with_log([Exception], message="Failed to get check presets data from cache"):
            cached_data = await self._cache_source.get_check_presets_data(
                access_key, group_name, domain_name, scaling_group
            )
            if cached_data:
                log.info(
                    "Cache hit for check_presets: {}, {}, {}", access_key, group_name, domain_name
                )
                return CheckPresetsResult.from_cache(cached_data)
        log.info(
            "Cache miss for check_presets, fetching from DB, {}, {}, {}",
            access_key,
            group_name,
            domain_name,
        )

        group_resource_visibility = await self._config_provider.legacy_etcd_config_loader.get_raw(
            "config/api/resources/group_resource_visibility"
        )
        group_resource_visibility = t.ToBool().check(group_resource_visibility)

        # Get all data from DB source
        db_data = await self._db_source.check_presets_data(
            access_key,
            user_id,
            group_name,
            domain_name,
            resource_policy,
            known_slot_types,
            scaling_group,
        )

        # Process the data and build response

        # Apply group resource visibility settings
        if not group_resource_visibility:
            nan_slots = ResourceSlot({k: Decimal("NaN") for k in db_data.known_slot_types.keys()})
            group_limits = nan_slots
            group_occupied = nan_slots
            group_remaining = nan_slots
        else:
            group_limits = db_data.keypair_data.group_limits
            group_occupied = db_data.keypair_data.group_occupied
            group_remaining = db_data.keypair_data.group_remaining

        result = CheckPresetsResult(
            presets=db_data.presets,
            keypair_limits=db_data.keypair_data.limits,
            keypair_using=db_data.keypair_data.occupied,
            keypair_remaining=db_data.keypair_data.remaining,
            group_limits=group_limits,
            group_using=group_occupied,
            group_remaining=group_remaining,
            scaling_group_remaining=db_data.keypair_data.scaling_group_remaining,
            scaling_groups=db_data.per_sgroup_data,
        )

        # Cache the result
        with suppress_with_log([Exception], message="Failed to cache check presets data"):
            cache_data = result.to_cache()
            await self._cache_source.set_check_presets_data(
                access_key,
                group_name,
                domain_name,
                scaling_group,
                cache_data,
            )

        return result



"""
Kernel and session-related exceptions.
"""

from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Optional, Union, cast

from aiohttp import web

from ai.backend.common.exception import (
    BackendAIError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)
from ai.backend.common.json import dump_json

from ..exceptions import AgentError
from .common import ObjectNotFound

if TYPE_CHECKING:
    pass


class KernelNotReady(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/kernel-not-ready"
    error_title = "Kernel not ready."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.KERNEL,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_READY,
        )


class SessionNotFound(ObjectNotFound):
    object_name = "session"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.SESSION,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class MainKernelNotFound(ObjectNotFound):
    object_name = "main kernel"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.KERNEL,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class KernelNotFound(ObjectNotFound):
    object_name = "kernel"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.KERNEL,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class TooManySessionsMatched(BackendAIError, web.HTTPNotFound):
    error_type = "https://api.backend.ai/probs/too-many-sessions-matched"
    error_title = "Too many sessions matched."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.SESSION,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.CONFLICT,
        )

    def __init__(
        self,
        extra_msg: Optional[str] = None,
        extra_data: Optional[dict[str, Any]] = None,
        **kwargs,
    ):
        if extra_data is not None and (matches := extra_data.get("matches", None)) is not None:
            serializable_matches = [
                {
                    "id": str(item["session_id"]),
                    "name": item["session_name"],
                    "status": item["status"].name,
                    "created_at": item["created_at"].isoformat(),
                }
                for item in matches
            ]
            extra_data["matches"] = serializable_matches
        super().__init__(extra_msg, extra_data, **kwargs)


class TooManyKernelsFound(BackendAIError, web.HTTPNotFound):
    error_type = "https://api.backend.ai/probs/too-many-kernels"
    error_title = "There are two or more matching kernels."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.KERNEL,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.CONFLICT,
        )


class SessionAlreadyExists(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/session-already-exists"
    error_title = "The session already exists but you requested not to reuse existing one."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.SESSION,
            operation=ErrorOperation.CREATE,
            error_detail=ErrorDetail.ALREADY_EXISTS,
        )


class QuotaExceeded(BackendAIError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/quota-exceeded"
    error_title = "You have reached your resource limit."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.SESSION,
            operation=ErrorOperation.CREATE,
            error_detail=ErrorDetail.UNAVAILABLE,
        )


class BackendAgentError(BackendAIError):
    """
    An RFC-7807 error class that wraps agent-side errors.
    """

    _short_type_map = {
        "TIMEOUT": "https://api.backend.ai/probs/agent-timeout",
        "INVALID_INPUT": "https://api.backend.ai/probs/agent-invalid-input",
        "FAILURE": "https://api.backend.ai/probs/agent-failure",
    }

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.AGENT,
            operation=ErrorOperation.ACCESS,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )

    def __init__(
        self,
        agent_error_type: str,
        exc_info: Union[str, AgentError, Exception, Mapping[str, Optional[str]], None] = None,
    ):
        super().__init__()
        agent_details: Mapping[str, Optional[str]]
        if not agent_error_type.startswith("https://"):
            agent_error_type = self._short_type_map[agent_error_type.upper()]
        self.args = (
            self.status_code,
            self.reason,
            self.error_type,
            agent_error_type,
        )
        if isinstance(exc_info, str):
            agent_details = {
                "type": agent_error_type,
                "title": exc_info,
            }
        elif isinstance(exc_info, AgentError):
            e = cast(AgentError, exc_info)
            agent_details = {
                "type": agent_error_type,
                "title": "Agent-side exception occurred.",
                "exception": e.exc_repr,
            }
        elif isinstance(exc_info, Exception):
            agent_details = {
                "type": agent_error_type,
                "title": "Unexpected exception ocurred.",
                "exception": repr(exc_info),
            }
        elif isinstance(exc_info, Mapping):
            agent_details = exc_info
        else:
            agent_details = {
                "type": agent_error_type,
                "title": None if exc_info is None else str(exc_info),
            }
        self.agent_details = agent_details
        self.agent_error_type = agent_error_type
        self.agent_error_title = agent_details["title"]
        self.agent_exception = agent_details.get("exception", "")
        self.body = dump_json({
            "type": self.error_type,
            "title": self.error_title,
            "agent-details": agent_details,
        })

    def __str__(self):
        if self.agent_exception:
            return f"{self.agent_error_title} ({self.agent_exception})"
        return f"{self.agent_error_title}"

    def __repr__(self):
        if self.agent_exception:
            return f"<{type(self).__name__}: {self.agent_error_title} ({self.agent_exception})>"
        return f"<{type(self).__name__}: {self.agent_error_title}>"

    def __reduce__(self):
        return (type(self), (self.agent_error_type, self.agent_details))


class KernelCreationFailed(BackendAgentError, web.HTTPInternalServerError):
    error_type = "https://api.backend.ai/probs/kernel-creation-failed"
    error_title = "Kernel creation has failed."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.KERNEL,
            operation=ErrorOperation.CREATE,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class KernelDestructionFailed(BackendAgentError, web.HTTPInternalServerError):
    error_type = "https://api.backend.ai/probs/kernel-destruction-failed"
    error_title = "Kernel destruction has failed."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.KERNEL,
            operation=ErrorOperation.SOFT_DELETE,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class KernelRestartFailed(BackendAgentError, web.HTTPInternalServerError):
    error_type = "https://api.backend.ai/probs/kernel-restart-failed"
    error_title = "Kernel restart has failed."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.KERNEL,
            operation=ErrorOperation.START,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class KernelExecutionFailed(BackendAgentError, web.HTTPInternalServerError):
    error_type = "https://api.backend.ai/probs/kernel-execution-failed"
    error_title = "Executing user code in the kernel has failed."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.KERNEL,
            operation=ErrorOperation.EXECUTE,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )



from collections.abc import Mapping
from dataclasses import dataclass
from datetime import datetime
from functools import lru_cache
from typing import Any, Optional
from uuid import UUID

from ai.backend.common.types import (
    CIStrEnum,
    ResourceSlot,
    SessionResult,
    SessionTypes,
)
from ai.backend.manager.data.image.types import ImageIdentifier


class KernelStatus(CIStrEnum):
    # values are only meaningful inside the manager
    PENDING = "PENDING"
    # ---
    SCHEDULED = "SCHEDULED"
    PREPARING = "PREPARING"
    # ---
    BUILDING = "BUILDING"
    PULLING = "PULLING"
    PREPARED = "PREPARED"
    CREATING = "CREATING"
    # ---
    RUNNING = "RUNNING"
    RESTARTING = "RESTARTING"
    RESIZING = "RESIZING"
    SUSPENDED = "SUSPENDED"
    # ---
    TERMINATING = "TERMINATING"
    TERMINATED = "TERMINATED"
    ERROR = "ERROR"
    CANCELLED = "CANCELLED"

    @classmethod
    def having_containers(cls) -> set["KernelStatus"]:
        return {
            cls.PULLING,
            cls.CREATING,
            cls.RUNNING,
        }

    def have_container(self) -> bool:
        """
        Check if the current status is one of the statuses that have containers.
        """
        return self in KernelStatus.having_containers()

    @classmethod
    @lru_cache(maxsize=1)
    def resource_occupied_statuses(cls) -> frozenset["KernelStatus"]:
        """
        Returns a set of kernel statuses that are considered as resource-occupying.
        """
        return frozenset((
            cls.RUNNING,
            cls.TERMINATING,
        ))

    @classmethod
    @lru_cache(maxsize=1)
    def resource_requested_statuses(cls) -> frozenset["KernelStatus"]:
        """
        Returns a set of kernel statuses that are considered as resource-occupying.
        """
        return frozenset((
            cls.SCHEDULED,
            cls.PREPARING,
            cls.PULLING,
            cls.PREPARED,
            cls.CREATING,
        ))

    @classmethod
    @lru_cache(maxsize=1)
    def terminatable_statuses(cls) -> frozenset["KernelStatus"]:
        """Return statuses that can transition to TERMINATING."""
        return frozenset(
            status
            for status in cls
            if status
            not in (
                cls.PENDING,
                cls.TERMINATING,
                cls.TERMINATED,
                cls.CANCELLED,
                cls.ERROR,
            )
        )

    @classmethod
    @lru_cache(maxsize=1)
    def terminal_statuses(cls) -> frozenset["KernelStatus"]:
        """
        Returns a set of kernel statuses that are considered terminal.
        """
        return frozenset((
            cls.ERROR,
            cls.TERMINATED,
            cls.CANCELLED,
        ))

    @classmethod
    @lru_cache(maxsize=1)
    def retriable_statuses(cls) -> frozenset["KernelStatus"]:
        """
        Returns a set of kernel statuses that are considered retriable.
        """
        return frozenset(
            (
                status
                for status in cls
                if status
                not in (
                    cls.RUNNING,
                    cls.TERMINATING,
                    cls.TERMINATED,
                    cls.CANCELLED,
                )
            )
        )


@dataclass
class RelatedSessionInfo:
    session_id: str  # Session UUID
    creation_id: Optional[str]
    name: Optional[str]
    session_type: SessionTypes


@dataclass
class ClusterConfig:
    cluster_mode: str
    cluster_size: int
    cluster_role: str
    cluster_idx: int
    local_rank: int
    cluster_hostname: str


@dataclass
class UserPermission:
    user_uuid: UUID
    access_key: str
    domain_name: str
    group_id: UUID
    uid: Optional[int]
    main_gid: Optional[int]
    gids: Optional[list[int]]


@dataclass
class ResourceInfo:
    scaling_group: Optional[str]
    agent: Optional[str]
    agent_addr: Optional[str]
    container_id: Optional[str]
    occupied_slots: ResourceSlot
    requested_slots: ResourceSlot
    occupied_shares: dict[str, Any]
    attached_devices: dict[str, Any]
    resource_opts: dict[str, Any]


@dataclass
class ImageInfo:
    identifier: Optional[ImageIdentifier]
    registry: Optional[str]
    tag: Optional[str]
    architecture: Optional[str]


@dataclass
class RuntimeConfig:
    environ: Optional[list[str]]
    mounts: Optional[list[str]]  # legacy
    mount_map: Optional[dict[str, Any]]  # legacy
    vfolder_mounts: Optional[list[dict[str, Any]]]
    bootstrap_script: Optional[str]
    startup_command: Optional[str]


@dataclass
class NetworkConfig:
    kernel_host: Optional[str]
    repl_in_port: int
    repl_out_port: int
    stdin_port: int  # legacy
    stdout_port: int  # legacy
    service_ports: Optional[dict[str, Any]]
    preopen_ports: Optional[list[int]]
    use_host_network: bool


@dataclass
class LifecycleStatus:
    status: KernelStatus
    result: SessionResult
    created_at: Optional[datetime]
    terminated_at: Optional[datetime]
    starts_at: Optional[datetime]
    status_changed: Optional[datetime]
    status_info: Optional[str]
    status_data: Optional[Mapping[str, Any]]
    status_history: Optional[dict[str, Any]]
    last_seen: Optional[datetime]


@dataclass
class Metrics:
    num_queries: int
    last_stat: Optional[dict[str, Any]]
    container_log: Optional[bytes]


@dataclass
class Metadata:
    callback_url: Optional[str]
    internal_data: Optional[dict[str, Any]]


@dataclass
class KernelInfo:
    id: UUID  # Kernel UUID
    session: RelatedSessionInfo
    user_permission: UserPermission
    image: ImageInfo
    network: NetworkConfig
    cluster: ClusterConfig
    resource: ResourceInfo
    runtime: RuntimeConfig
    lifecycle: LifecycleStatus
    metrics: Metrics
    metadata: Metadata



from dataclasses import dataclass
from typing import Any, Optional, override

from ai.backend.common.types import AccessKey
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.session.types import SessionData
from ai.backend.manager.services.session.base import SessionAction


@dataclass
class ExecuteSessionActionParams:
    mode: Optional[str]
    # TODO: Add proper type
    options: Optional[Any]
    code: Optional[str]
    run_id: Optional[str]


@dataclass
class ExecuteSessionAction(SessionAction):
    session_name: str
    api_version: tuple[Any, ...]
    owner_access_key: AccessKey
    params: ExecuteSessionActionParams

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "execute"


@dataclass
class ExecuteSessionActionResult(BaseActionResult):
    # TODO: Add proper type
    result: Any
    session_data: SessionData

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.session_data.id)
